import uuid
from collections import deque
from itertools import zip_longest

import pytest

from bluesky import Msg
from bluesky.preprocessors import (
    finalize_wrapper,
    msg_mutator,
    pchain,
    plan_mutator,
    stub_wrapper,
)
from bluesky.preprocessors import (
    single_gen as single_message_gen,
)
from bluesky.utils import ensure_generator


class EchoException(Exception): ...


def EchoRE(plan, *, debug=False, msg_list=None):
    """An 'echo' RunEngine for testing.

    Always sends the message back into the plan as the result.

    Parameters
    ----------
    plan : iterable
        The plan to run through

    debug : bool, optional (False)
        print the messages on the way by

    msg_list : mutable sequence, optional
        If not None, mutate this object by appending messages.
        This is the easiest way to capture messages if the plan
        raises.

    Returns
    -------
    msg_list : list
        List of all the messages seen by the RE
    """
    if msg_list is None:
        msg_list = deque()
    ret = None
    plan = ensure_generator(plan)
    while True:
        try:
            msg = plan.send(ret)
            if debug:
                print(msg)
            msg_list.append(msg)
            if msg.command == "FAIL":
                plan.throw(EchoException(msg))
            ret = msg
        except StopIteration:
            break

    return list(msg_list)


def echo_plan(*, command="echo", num=4):
    """Testing plan which expects to get back a message with an equal object"""
    seed = str(uuid.uuid4())[:6]
    for ch in map(lambda x: chr(97 + x), range(num)):  # noqa: C417
        sent = f"{seed}_{ch}"
        ret = yield Msg(command, sent)
        assert ret.obj == sent


def test_mutator_exceptions():
    handled = False

    def base_plan():
        yield Msg("foo")
        yield Msg("bar")

    def failing_plan():
        nonlocal handled
        handled = False
        yield Msg("pre")
        try:
            yield Msg("FAIL")
        except EchoException:
            handled = True
            raise

    def test_mutator(msg):
        if msg.command == "bar":
            return (failing_plan(), single_message_gen(Msg("foo")))
        return None, None

    # check generator exit behavior
    plan = plan_mutator(base_plan(), test_mutator)
    next(plan)
    plan.close()

    # check exception fall through
    plan = plan_mutator(base_plan(), test_mutator)
    with pytest.raises(EchoException):
        EchoRE(plan, debug=True)
    assert handled


def _verify_msg_seq(msgs, *, cmd_sq=None, obj_sq=None, args_sq=None, kwargs_sq=None, m_len=None):
    def _verify_cmpt(msgs, seq, cmpt):
        for m, s in zip_longest(msgs, seq):
            assert getattr(m, cmpt) == s

    if m_len is not None:
        assert len(msgs) == m_len

    if cmd_sq is not None:
        _verify_cmpt(msgs, cmd_sq, "command")

    if obj_sq is not None:
        _verify_cmpt(msgs, obj_sq, "obj")

    if args_sq is not None:
        _verify_cmpt(msgs, args_sq, "args")

    if kwargs_sq is not None:
        _verify_cmpt(msgs, kwargs_sq, "kwargs")


def test_smoke_test():
    num = 10
    cmd = "smoke"
    msgs = EchoRE(echo_plan(command=cmd, num=num))
    _verify_msg_seq(msgs, m_len=num, cmd_sq=[cmd] * num, args_sq=[()] * num, kwargs_sq=[{}] * num)


def test_simple_replace():
    new_cmd = "replaced"

    def change_command(msg):
        return msg._replace(command=new_cmd)

    num = 10
    msgs = EchoRE(msg_mutator(echo_plan(num=num), change_command))
    _verify_msg_seq(msgs, m_len=num, cmd_sq=[new_cmd] * num, args_sq=[()] * num, kwargs_sq=[{}] * num)


def test_simple_mutator():
    _mut_active = True
    pre_count = 3
    post_count = 5
    pre_cmd = "pre"
    post_cmd = "post"

    def test_mutator(msg):
        nonlocal _mut_active
        if _mut_active:
            _mut_active = False

            return (
                pchain(echo_plan(num=pre_count, command=pre_cmd), single_message_gen(msg)),
                echo_plan(num=post_count, command=post_cmd),
            )
        return None, None

    num = 5
    cmd = "echo"
    plan = plan_mutator(echo_plan(command=cmd, num=num), test_mutator)
    msgs = EchoRE(plan)
    total = num + pre_count + post_count
    cmd_sq = [pre_cmd] * pre_count + [cmd] + [post_cmd] * post_count + [cmd] * (num - 1)
    _verify_msg_seq(msgs, m_len=total, cmd_sq=cmd_sq, args_sq=[()] * total, kwargs_sq=[{}] * total)


def test_finialize_fail():
    fail_cmd = "fail_next"

    def erroring_plan():
        yield Msg(fail_cmd, None)
        raise RuntimeError("saw this coming")

    num = 5
    cmd = "echo"
    plan = finalize_wrapper(erroring_plan(), echo_plan(command=cmd, num=num))
    msgs = list()  # noqa: C408
    try:
        EchoRE(plan, msg_list=msgs)
    except RuntimeError:
        pass

    total = num + 1
    _verify_msg_seq(
        msgs, m_len=total, cmd_sq=[fail_cmd] + [cmd] * num, args_sq=[()] * total, kwargs_sq=[{}] * total
    )


def test_finialize_pause():
    fail_cmd = "fail_next"

    def erroring_plan():
        yield Msg(fail_cmd, None)
        raise RuntimeError("saw this coming")

    num = 5
    cmd = "echo"
    plan = finalize_wrapper(erroring_plan(), echo_plan(command=cmd, num=num), pause_for_debug=True)
    msgs = list()  # noqa: C408
    try:
        EchoRE(plan, msg_list=msgs)
    except RuntimeError:
        pass

    total = num + 2
    _verify_msg_seq(
        msgs,
        m_len=total,
        cmd_sq=[fail_cmd, "pause"] + [cmd] * num,
        args_sq=[()] * total,
        kwargs_sq=[{}, {"defer": False}] + [{}] * num,
    )


def test_finialize_success():
    suc_cmd = "it_works"

    num = 5
    cmd = "echo"
    plan = finalize_wrapper(single_message_gen(Msg(suc_cmd, None)), echo_plan(command=cmd, num=num))

    msgs = list()  # noqa: C408
    try:
        EchoRE(plan, msg_list=msgs)
    except RuntimeError:
        pass

    total = num + 1
    _verify_msg_seq(
        msgs, m_len=total, cmd_sq=[suc_cmd] + [cmd] * num, args_sq=[()] * total, kwargs_sq=[{}] * total
    )


def test_plan_mutator_exception_propogation():
    class ExpectedException(Exception):
        pass

    num = 5
    cmd1 = "echo1"
    cmd2 = "echo2"

    def bad_tail():
        yield Msg("one_tail", None)
        raise ExpectedException("this is a test")

    def sarfing_plan():
        try:
            yield from echo_plan(command=cmd1, num=num)
        except ExpectedException:
            print("CAUGHT IT")

    _mut_active = True

    def test_mutator(msg):
        nonlocal _mut_active
        if _mut_active:
            _mut_active = False

            return (pchain(echo_plan(num=2, command=cmd2), single_message_gen(msg)), bad_tail())
        return None, None

    plan = plan_mutator(sarfing_plan(), test_mutator)
    EchoRE(plan, debug=True)


def test_exception_in_pre_with_tail():
    class SnowFlake(Exception): ...

    def bad_pre():
        yield Msg("pre_bad", None)
        raise SnowFlake("this one")

    def good_post():
        yield Msg("good_post", None)

    def test_mutator(msg):
        if msg.command == "TARGET":
            return bad_pre(), good_post()

        return None, None

    def testing_plan():
        yield Msg("a", None)
        yield Msg("b", None)
        try:
            yield Msg("TARGET", None)
        except SnowFlake:
            pass
        yield Msg("b", None)
        yield Msg("a", None)

    plan = plan_mutator(testing_plan(), test_mutator)
    msgs = EchoRE(plan, debug=True)
    _verify_msg_seq(msgs, m_len=5, cmd_sq=["a", "b", "pre_bad", "b", "a"], args_sq=[()] * 5, kwargs_sq=[{}] * 5)


def test_plan_mutator_returns():
    def testing_plan():
        yield Msg("a", None)
        yield Msg("TARGET", None)
        yield Msg("b", None)

        return "foobar"

    def outer_plan(pln):
        ret = yield from pln
        assert ret == "foobar"
        return ret

    def tail_plan():
        yield Msg("A", None)
        return "baz"

    def test_mutator(msg):
        def pre_plan():
            yield Msg("pre", None)
            yield msg

        if msg.command == "TARGET":
            return pre_plan(), tail_plan()

        return None, None

    plan = plan_mutator(testing_plan(), test_mutator)
    msgs = EchoRE(plan)
    _verify_msg_seq(msgs, m_len=5, cmd_sq=["a", "pre", "TARGET", "A", "b"], args_sq=[()] * 5, kwargs_sq=[{}] * 5)


def test_insert_before():
    def target():
        yield Msg("a", None)
        ret = yield Msg("TARGET", None)
        yield Msg("b", None)
        assert ret.command == "TARGET"
        return ret
        return ret

    def insert_before(msg):
        if msg.command == "TARGET":

            def pre():
                yield Msg("pre", None)
                ret = yield msg
                assert ret is not None
                assert ret.command == "TARGET"
                return ret

            return pre(), None
        else:
            return None, None

    EchoRE(plan_mutator(target(), insert_before))


def test_insert_after():
    def target():
        yield Msg("a", None)
        ret = yield Msg("TARGET", None)
        yield Msg("b", None)
        assert ret is not None
        assert ret.command == "TARGET"
        return ret

    def insert_after(msg):
        if msg.command == "TARGET":

            def post():
                yield Msg("post", None)

            return None, post()
        else:
            return None, None

    EchoRE(plan_mutator(target(), insert_after))


def test_base_exception():
    class SnowFlake(Exception): ...

    def null_mutator(msg):
        return None, None

    def test_plan():
        yield Msg("a", None)
        raise SnowFlake("this one")

    pln = plan_mutator(test_plan(), null_mutator)

    try:
        EchoRE(pln)
    except SnowFlake as ex:
        assert ex.args[0] == "this one"


def test_msg_mutator_skip():
    def skipper(msg):
        if msg.command == "SKIP":
            return None
        return msg

    def skip_plan():
        for c in "abcd":
            yield Msg(c, None)
            yield Msg("SKIP", None)

    pln = msg_mutator(skip_plan(), skipper)
    msgs = EchoRE(pln)
    _verify_msg_seq(msgs, m_len=4, cmd_sq="abcd", args_sq=[()] * 4, kwargs_sq=[{}] * 4)


def test_stub_wrapper():
    def plan():
        yield Msg("open_run")
        yield Msg("stage")
        yield Msg("read")
        yield Msg("unstage")
        yield Msg("close_run")

    stub_plan = list(stub_wrapper(plan()))
    assert len(stub_plan) == 1
    assert stub_plan[0].command == "read"



"""
Best Effort Callback.
For instructions on how to test in a simulated environment please see:
    tests/interactive/best_effort_cb.py
"""

import itertools
import logging
import re
import sys
import threading
import time
import weakref
from collections.abc import Mapping
from datetime import datetime
from functools import partial
from io import StringIO
from pprint import pformat
from typing import Any, Literal
from warnings import warn

import matplotlib.pyplot as plt
import numpy as np
from cycler import cycler
from matplotlib.axes import Axes
from matplotlib.axis import Axis
from matplotlib.figure import Figure

from .core import LiveTable, make_class_safe
from .fitting import PeakStats
from .mpl_plotting import LiveGrid, LivePlot, LiveScatter, QtAwareCallback

logger = logging.getLogger(__name__)


@make_class_safe(logger=logger)
class BestEffortCallback(QtAwareCallback):
    def __init__(self, *, fig_factory=None, table_enabled=True, calc_derivative_and_stats=False, **kwargs):
        super().__init__(**kwargs)
        # internal state
        self._start_doc = None
        self._descriptors = {}
        self._table = None
        self._heading_enabled = True
        self._table_enabled = table_enabled
        self._baseline_enabled = True
        self._plots_enabled = True
        # axes supplied from outside
        self._fig_factory = fig_factory if fig_factory is not None else partial(plt.figure, layout="constrained")
        # maps descriptor uid to dict which maps data key to LivePlot instance
        self._live_plots = {}
        self._live_grids = {}
        self._live_scatters = {}
        self._peak_stats = {}  # same structure as live_plots
        self._calc_derivative_and_stats = calc_derivative_and_stats
        self._cleanup_motor_heuristic = False
        self._stream_names_seen = set()

        # public options
        self.overplot = True
        self.noplot_streams = ["baseline"]
        self.omit_single_point_plot = True

        # public data
        self.peaks = PeakResults()

        # hack to handle the bottom border of the table
        self._buffer = StringIO()
        self._baseline_toggle = True

    def enable_heading(self):
        "Print timestamp and IDs at the top of a run."
        self._heading_enabled = True

    def disable_heading(self):
        "Opposite of enable_heading()"
        self._heading_enabled = False

    def enable_table(self):
        "Print hinted readings from the 'primary' stream in a LiveTable."
        self._table_enabled = True

    def disable_table(self):
        "Opposite of enable_table()"
        self._table_enabled = False

    def enable_baseline(self):
        "Print hinted fields from the 'baseline' stream."
        self._baseline_enabled = True

    def disable_baseline(self):
        "Opposite of enable_baseline()"
        self._baseline_enabled = False

    def enable_plots(self):
        "Plot hinted fields from all streams not in ``noplot_streams``."
        self._plots_enabled = True

    def disable_plots(self):
        "Do not plot anything."
        self._plots_enabled = False

    def __call__(self, name, doc, *args, **kwargs):
        if not (self._table_enabled or self._baseline_enabled or self._plots_enabled):
            return
        super().__call__(name, doc, *args, **kwargs)

    def start(self, doc):
        self.clear()
        self._start_doc = doc
        self.plan_hints = doc.get("hints", {})

        # Prepare a guess about the dimensions (independent variables) in case
        # we need it.
        motors = self._start_doc.get("motors")
        if motors is not None:
            GUESS = [([motor], "primary") for motor in motors]
        else:
            GUESS = [(["time"], "primary")]

        # Ues the guess if there is not hint about dimensions.
        dimensions = self.plan_hints.get("dimensions")
        if dimensions is None:
            self._cleanup_motor_heuristic = True
            dimensions = GUESS

        # We can only cope with all the dimensions belonging to the same
        # stream unless we resample. We are not doing to handle that yet.
        if len(set(d[1] for d in dimensions)) != 1:  # noqa: C401
            self._cleanup_motor_heuristic = True
            dimensions = GUESS  # Fall back on our GUESS.
            warn(
                "We are ignoring the dimensions hinted because we cannot combine streams.",
                stacklevel=1,
            )

        # for each dimension, choose one field only
        # the plan can supply a list of fields. It's assumed the first
        # of the list is always the one plotted against
        default_field_name = "time"
        self.dim_fields = [fields[0] if len(fields) else default_field_name for fields, stream_name in dimensions]

        # make distinction between flattened fields and plotted fields
        # motivation for this is that when plotting, we find dependent variable
        # by finding elements that are not independent variables
        self.all_dim_fields = [field for fields, stream_name in dimensions for field in fields]

        _, self.dim_stream = dimensions[0]

        # Print heading.
        tt = datetime.fromtimestamp(self._start_doc["time"]).utctimetuple()
        if self._heading_enabled:
            print(
                "\n\nTransient Scan ID: {0}     Time: {1}".format(  # noqa: UP030
                    self._start_doc.get("scan_id", ""), time.strftime("%Y-%m-%d %H:%M:%S", tt)
                )
            )
            print("Persistent Unique Scan ID: '{0}'".format(self._start_doc["uid"]))  # noqa: UP030

    def _set_up_plots(self, doc, stream_name, columns: list[Any]):
        """Using the descriptor doc"""
        plot_data = self._verify_plot_data(stream_name, columns)
        if not plot_data:
            return

        # This is a heuristic approach until we think of how to hint this in a
        # generalizable way.
        # 'time' once LivePlot can do that
        dim_fields: list[str] = self.dim_fields if stream_name == self.dim_stream else ["time"]

        ndims = len(dim_fields)
        if not 0 < ndims < 3:
            # we need 1 or 2 dims to do anything, do not make empty figures
            warn(  # noqa: B028
                "Plots are only made for 1 or 2 dimensions. "
                "Adjust the metadata hints field for BestEffortCallback to produce plots."
            )
            return
        # Create a figure or reuse an existing one.

        figure_name = self._get_figure_name(columns, dim_fields)

        figure = self._fig_factory(figure_name)

        axes = self._generate_axes(columns, ndims, figure)

        # ## LIVE PLOT AND PEAK ANALYSIS ## #

        if ndims == 1:
            self._live_plots[doc["uid"]] = {}
            self._peak_stats[doc["uid"]] = {}
            (x_key,) = dim_fields
            for y_key, ax in zip(columns, axes):
                dtype = doc["data_keys"][y_key]["dtype"]
                if dtype not in ("number", "integer"):
                    warn(
                        f"Omitting {y_key} from plot because dtype is {dtype}",
                        stacklevel=1,
                    )
                    continue
                # Create an instance of LivePlot and an instance of PeakStats.
                live_plot = LivePlotPlusPeaks(y=y_key, x=x_key, ax=ax, peak_results=self.peaks)
                live_plot("start", self._start_doc)
                live_plot("descriptor", doc)
                peak_stats = PeakStats(x=x_key, y=y_key, calc_derivative_and_stats=self._calc_derivative_and_stats)
                peak_stats("start", self._start_doc)
                peak_stats("descriptor", doc)

                # Stash them in state.
                self._live_plots[doc["uid"]][y_key] = live_plot
                self._peak_stats[doc["uid"]][y_key] = peak_stats

            for ax in axes[:-1]:
                ax.set_xlabel("")
        elif ndims == 2:
            # Decide whether to use LiveGrid or LiveScatter. LiveScatter is the
            # safer one to use, so it is the fallback..
            gridding = self._start_doc.get("hints", {}).get("gridding")
            if gridding == "rectilinear":
                self._live_grids[doc["uid"]] = {}
                slow, fast = dim_fields
                try:
                    extents = self._start_doc["extents"]
                    shape = self._start_doc["shape"]
                except KeyError:
                    warn(
                        "Need both 'shape' and 'extents' in plan metadata to create LiveGrid.",
                        stacklevel=1,
                    )
                else:
                    data_range = np.array([float(e[-1] - e[0]) for e in extents])
                    y_step, x_step = data_range / [max(1, s - 1) for s in shape]
                    adjusted_extent = [
                        extents[1][0] - x_step / 2,
                        extents[1][1] + x_step / 2,
                        extents[0][0] - y_step / 2,
                        extents[0][1] + y_step / 2,
                    ]
                    # MAGIC NUMBERS based on what tacaswell thinks looks OK
                    ASPECT_RATIO_REF_VALUE = 2
                    for I_key, ax in zip(columns, axes):
                        data_aspect_ratio = np.abs(data_range[1] / data_range[0])
                        aspect: Literal["auto", "equal"]
                        if 1 / ASPECT_RATIO_REF_VALUE < data_aspect_ratio < ASPECT_RATIO_REF_VALUE:
                            aspect = "equal"
                            ax.set_aspect(aspect, adjustable="box")
                        else:
                            aspect = "auto"
                            ax.set_aspect(aspect, adjustable="datalim")

                        live_grid = LiveGrid(
                            shape,
                            I_key,
                            xlabel=fast,
                            ylabel=slow,
                            extent=adjusted_extent,
                            aspect=aspect,
                            ax=ax,
                        )

                        live_grid("start", self._start_doc)
                        live_grid("descriptor", doc)
                        self._live_grids[doc["uid"]][I_key] = live_grid
            else:
                self._live_scatters[doc["uid"]] = {}
                x_key, y_key = dim_fields
                for I_key, ax in zip(columns, axes):
                    try:
                        extents = self._start_doc["extents"]
                    except KeyError:
                        xlim = ylim = None
                    else:
                        xlim, ylim = extents
                    live_scatter = LiveScatter(
                        x_key,
                        y_key,
                        I_key,
                        xlim=xlim,
                        ylim=ylim,
                        # Let clim autoscale.
                        ax=ax,
                    )
                    live_scatter("start", self._start_doc)
                    live_scatter("descriptor", doc)
                    self._live_scatters[doc["uid"]][I_key] = live_scatter

        else:
            raise NotImplementedError("we do not support 3D+ in BEC yet (and it should have bailed above)")

    def _generate_axes(self, columns, ndims, figure: Figure) -> list[Axes]:
        if not figure.axes:
            if len(columns) < 5:
                layout = (len(columns), 1)
            else:
                nrows = ncols = int(np.ceil(np.sqrt(len(columns))))
                while (nrows - 1) * ncols >= len(columns):
                    nrows -= 1
                layout = (nrows, ncols)
            if ndims == 1:
                share_kwargs = {"sharex": True}
            elif ndims == 2:
                share_kwargs = {"sharex": True, "sharey": True}
            else:
                raise NotImplementedError("we now support 3D?!")

            fig_size = np.array(layout[::-1]) * 5
            figure.set_size_inches(*fig_size)
            axes_grid = figure.subplots(*map(int, layout), **share_kwargs)  # type: ignore
            for ax in figure.axes[len(columns) :]:
                ax.set_visible(False)

            if len(figure.axes) > 1 and len(axes_grid.shape) == 2:
                # Axes go left to right, top to bottom, and will make some labels invisible
                for i in range(int(layout[1])):
                    if axes_grid[-1, i].get_visible() is False:
                        axes_grid[-2, i].tick_params(axis="x", labelbottom=True)

        axes = figure.axes
        return axes

    def _get_figure_name(self, columns, dim_fields: list[str]) -> str:
        fig_name = "{} vs {}".format(" ".join(sorted(columns)), " ".join(sorted(dim_fields)))
        if self.overplot and len(dim_fields) == 1:
            # If any open figure matches 'figname {number}', use it. If there
            # are multiple, the most recently touched one will be used.
            pattern1 = re.compile("^" + fig_name + "$")
            pattern2 = re.compile("^" + fig_name + r" \d+$")
            for label in plt.get_figlabels():
                if pattern1.match(label) or pattern2.match(label):
                    fig_name = label
                    break
        else:
            if plt.fignum_exists(fig_name):
                # Generate a unique name by appending a number.
                for number in itertools.count(2):
                    new_name = f"{fig_name} {number}"
                    if not plt.fignum_exists(new_name):
                        fig_name = new_name
                        break
        return fig_name

    def _verify_plot_data(self, stream_name, columns: list[Any]) -> bool:
        plot_data = True

        if not self._plots_enabled:
            plot_data = False
        if stream_name in self.noplot_streams:
            plot_data = False
        if not columns:
            plot_data = False
        if (
            (self._start_doc.get("num_points") == 1)
            and (stream_name == self.dim_stream)
            and self.omit_single_point_plot
        ):
            plot_data = False
        return plot_data

    def descriptor(self, doc):
        self._descriptors[doc["uid"]] = doc
        stream_name = doc.get("name", "primary")  # fall back for old docs

        if stream_name not in self._stream_names_seen:
            self._stream_names_seen.add(stream_name)
            if self._table_enabled:
                print(f"New stream: {stream_name!r}")

        columns = hinted_fields(doc)

        # ## This deals with old documents. ## #
        if stream_name == "primary" and self._cleanup_motor_heuristic:
            # We stashed object names in self.dim_fields, which we now need to
            # look up the actual fields for.
            self._cleanup_motor_heuristic = False
            fixed_dim_fields = []
            for obj_name in self.dim_fields:
                # Special case: 'time' can be a dim_field, but it's not an
                # object name. Just add it directly to the list of fields.
                if obj_name == "time":
                    fixed_dim_fields.append("time")
                    continue
                try:
                    fields = doc.get("hints", {}).get(obj_name, {})["fields"]
                except KeyError:
                    fields = doc["object_keys"][obj_name]
                fixed_dim_fields.extend(fields)
            self.dim_fields = fixed_dim_fields

        # Ensure that no independent variables ('dimensions') are
        # duplicated here.
        columns = [c for c in columns if c not in self.all_dim_fields]

        # ## DECIDE WHICH KIND OF PLOT CAN BE USED ## #
        self._set_up_plots(doc, stream_name, columns)

        # ## TABLE ## #
        if stream_name == self.dim_stream:
            if self._table_enabled:
                # plot everything, independent or dependent variables
                self._table = LiveTable(list(self.all_dim_fields) + columns, separator_lines=False)
                self._table("start", self._start_doc)
                self._table("descriptor", doc)

    def event(self, doc):
        descriptor = self._descriptors[doc["descriptor"]]
        if descriptor.get("name") == "primary":
            if self._table is not None:
                self._table("event", doc)

        # Show the baseline readings.
        if descriptor.get("name") == "baseline":
            columns = hinted_fields(descriptor)
            self._baseline_toggle = not self._baseline_toggle
            if self._baseline_toggle:
                file = self._buffer
                subject = "End-of-run"
            else:
                file = sys.stdout
                subject = "Start-of-run"
            if self._baseline_enabled:
                print(f"{subject} baseline readings:", file=file)
                border = "+" + "-" * 32 + "+" + "-" * 32 + "+"
                print(border, file=file)
                for k, v in doc["data"].items():
                    if k not in columns:
                        continue
                    print(f"| {k:>30} | {v:<30} |", file=file)
                print(border, file=file)

        for y_key in doc["data"]:
            live_plot = self._live_plots.get(doc["descriptor"], {}).get(y_key)
            if live_plot is not None:
                live_plot("event", doc)
            live_grid = self._live_grids.get(doc["descriptor"], {}).get(y_key)
            if live_grid is not None:
                live_grid("event", doc)
            live_sc = self._live_scatters.get(doc["descriptor"], {}).get(y_key)
            if live_sc is not None:
                live_sc("event", doc)
            peak_stats = self._peak_stats.get(doc["descriptor"], {}).get(y_key)
            if peak_stats is not None:
                peak_stats("event", doc)

    def stop(self, doc):
        if self._table is not None:
            self._table("stop", doc)

        # Compute peak stats and build results container.
        ps_by_key = {}  # map y_key to PeakStats instance
        for peak_stats in self._peak_stats.values():
            for y_key, ps in peak_stats.items():
                ps("stop", doc)
                ps_by_key[y_key] = ps
        self.peaks.update(ps_by_key)

        for live_plots in self._live_plots.values():
            for live_plot in live_plots.values():
                live_plot("stop", doc)

        for live_grids in self._live_grids.values():
            for live_grid in live_grids.values():
                live_grid("stop", doc)

        for live_scatters in self._live_scatters.values():
            for live_scatter in live_scatters.values():
                live_scatter("stop", doc)

        if self._baseline_enabled:
            # Print baseline below bottom border of table.
            self._buffer.seek(0)
            print(self._buffer.read())
            print("\n")

    def clear(self):
        self._start_doc = None
        self._descriptors.clear()
        self._stream_names_seen.clear()
        self._table = None
        self._live_plots.clear()
        self._peak_stats.clear()
        self._live_grids.clear()
        self._live_scatters.clear()
        self.peaks.clear()
        self._buffer = StringIO()
        self._baseline_toggle = True

    def plot_prune_fifo(self, num_lines, x_signal, y_signal):
        """
        Find the plot with axes x_signal and y_signal.  Replot with only the last *num_lines* lines.

        Example to remove all scans but the last:
        >>> bec.plot_prune_fifo(1, m1, noisy)

        Parameters
        ----------
        num_lines: int
            number of lines (plotted scans) to keep, must be >= 0
        x_signal: object
            instance of ophyd.Signal (or subclass),
            independent (x) axis
        y_signal: object
            instance of ophyd.Signal (or subclass),
            dependent (y) axis
        """
        if num_lines < 0:
            emsg = f"Argument 'num_lines' (given as {num_lines}) must be >= 0."
            raise ValueError(emsg)
        for liveplot in self._live_plots.values():
            lp = liveplot.get(y_signal.name)
            if lp is None or lp.x != x_signal.name or lp.y != y_signal.name:
                continue

            # pick out only the lines that contain plot data,
            # skipping the lines that show peak centers
            lines = [
                tr
                for tr in lp.ax.lines
                if len(tr._x) != 2 or len(tr._y) != 2 or (len(tr._x) == 2 and tr._x[0] != tr._x[1])
            ]
            if len(lines) > num_lines:
                keepers = lines[-num_lines:]
                for tr in lines:
                    if tr not in keepers:
                        tr.remove()
                lp.ax.legend()
                if num_lines > 0:
                    lp.update_plot()


class PeakResults:
    ATTRS = ("com", "cen", "max", "min", "fwhm")

    def __init__(self):
        for attr in self.ATTRS:
            setattr(self, attr, {})

    def clear(self):
        for attr in self.ATTRS:
            getattr(self, attr).clear()

    def update(self, ps_dict):
        for y_key, ps in ps_dict.items():
            for attr in self.ATTRS:
                getattr(self, attr)[y_key] = getattr(ps, attr)

    def __getitem__(self, key):
        if key in self.ATTRS:
            return getattr(self, key)
        raise KeyError(f"Keys are: {self.ATTRS}")

    def __repr__(self):
        # This is a proper eval-able repr, but with some manually-tweaked
        # whitespace to make it easier to parse.
        lines = []
        lines.append("{")
        for attr in self.ATTRS:
            lines.append(f"'{attr}':")
            for line in pformat(getattr(self, attr), width=1).split("\n"):
                lines.append(f"    {line}")
            lines.append(",")
        lines.append("}")
        return "\n".join(lines)


class LivePlotPlusPeaks(LivePlot):
    # Track state of axes, which may share instances of LivePlotPlusPeaks.
    __labeled: Mapping[Axis, bool] = weakref.WeakKeyDictionary()  # map ax to True/False
    __visible: Mapping[Axis, bool] = weakref.WeakKeyDictionary()  # map ax to True/False
    __instances: Mapping[Axis, list["LivePlotPlusPeaks"]] = (
        weakref.WeakKeyDictionary()
    )  # map ax to list of instances

    def __init__(self, *args, peak_results, **kwargs):
        self.__setup_lock = threading.Lock()
        self.__setup_event = threading.Event()
        super().__init__(*args, **kwargs)
        self.peak_results = peak_results

        def setup():
            # Run this code in start() so that it runs on the correct thread.
            with self.__setup_lock:
                if self.__setup_event.is_set():
                    return
                self.__setup_event.set()
            ax = self.ax  # for brevity
            if ax not in self.__visible:
                # This is the first instance of LivePlotPlusPeaks on these axes.
                # Set up matplotlib event handling.

                self.__visible[ax] = False

                def toggle(event):
                    if event.key == "P":
                        self.__visible[ax] = ~self.__visible[ax]
                        for instance in self.__instances[ax]:
                            instance.check_visibility()

                ax.figure.canvas.mpl_connect("key_press_event", toggle)

            if ax not in self.__instances:
                self.__instances[ax] = []
            self.__instances[ax].append(self)
            self.__arts = None

        self.__setup = setup

    def check_visibility(self):
        if self.__visible[self.ax]:
            if self.__arts is None:
                self.plot_annotations()
            else:
                for artist in self.__arts:
                    artist.set_visible(True)
        elif self.__arts is not None:
            for artist in self.__arts:
                artist.set_visible(False)
        self.ax.legend(loc="best")
        self.ax.figure.canvas.draw_idle()

    def plot_annotations(self):
        styles = iter(cycler("color", "kr"))
        vlines = []
        for style, attr in zip(styles, ["cen", "com"]):
            val = self.peak_results[attr][self.y]
            # Only put labels in this legend once per axis.
            if self.ax in self.__labeled:
                label = "_no_legend_"
            else:
                label = attr
            vlines.append(self.ax.axvline(val, label=label, **style))
        self.__labeled[self.ax] = None
        self.__arts = vlines

    def start(self, doc):
        super().start(doc)
        self.__setup()

    def stop(self, doc):
        self.check_visibility()
        super().stop(doc)


def hinted_fields(descriptor):
    # Figure out which columns to put in the table.
    obj_names = list(descriptor["object_keys"])
    # We will see if these objects hint at whether
    # a subset of their data keys ('fields') are interesting. If they
    # did, we'll use those. If these didn't, we know that the RunEngine
    # *always* records their complete list of fields, so we can use
    # them all unselectively.
    columns = []
    for obj_name in obj_names:
        try:
            fields = descriptor.get("hints", {}).get(obj_name, {})["fields"]
        except KeyError:
            fields = descriptor["object_keys"][obj_name]
        columns.extend(fields)
    return columns



from collections.abc import Generator

from opentelemetry.trace import get_current_span

from bluesky.plan_stubs import sleep
from bluesky.tracing import trace_plan, tracer


@trace_plan(tracer, "test_plan")
def _test_plan():
    yield from sleep(0)


def test_trace_plan_wrapper_gives_back_generator():
    assert isinstance(_test_plan(), Generator)


def test_trace_plan_wrapper_opens_span_but_doesnt_do_anything_without_trace_provider(RE):
    @trace_plan(tracer, "test_plan_2")
    def test_plan_2():
        yield from sleep(0)
        assert not get_current_span().is_recording()

    RE(test_plan_2())



import itertools
import operator
import time
import typing
import uuid
import warnings
from collections.abc import Awaitable, Callable, Hashable, Iterable, Mapping, Sequence
from functools import reduce
from typing import Any, Literal, Optional, Union

from cycler import cycler

from bluesky.suspenders import SuspenderBase

try:
    # cytools is a drop-in replacement for toolz, implemented in Cython
    from cytools import partition
except ImportError:
    from toolz import partition

from event_model import ComposeEvent
from event_model.documents import EventDescriptor

from .protocols import (
    Configurable,
    Flyable,
    Locatable,
    Location,
    Movable,
    PartialEvent,
    Preparable,
    Readable,
    Reading,
    Stageable,
    Status,
    Stoppable,
    Triggerable,
    check_supports,
)
from .utils import (
    CustomPlanMetadata,
    Msg,
    MsgGenerator,
    ScalarOrIterableFloat,
    all_safe_rewind,
    ensure_generator,
    get_hinted_fields,
    merge_cycler,
    plan,
    separate_devices,
    short_uid,
)
from .utils import (
    short_uid as _short_uid,
)

#: Any plan function that takes a reading given a list of Readables
TakeReading = Callable[[Sequence[Readable]], MsgGenerator[Mapping[str, Reading]]]


@plan
def declare_stream(
    *objs: Readable, name: str, collect: bool = False
) -> MsgGenerator[tuple[EventDescriptor, ComposeEvent]]:
    """
    Bundle future readings into a new Event document.

    Parameters
    ----------
    objs :
        objects whose readings will be present in the stream
    name : string, optional
        name given to event stream, used for convenient identification
        default is 'primary'
    collect : bool, optional
        collect as well as describe when declaring the stream
        default is `False`

    Yields
    ------
    msg : Msg
        Msg('create', name=name)

    See Also
    --------
    :func:`bluesky.plan_stubs.save`
    """
    return (yield Msg("declare_stream", None, *separate_devices(objs), name=name, collect=collect))


@plan
def create(name: str = "primary") -> MsgGenerator:
    """
    Bundle future readings into a new Event document.

    Parameters
    ----------
    name : string, optional
        name given to event stream, used for convenient identification
        default is 'primary'

    Yields
    ------
    msg : Msg
        Msg('create', name=name)

    See Also
    --------
    :func:`bluesky.plan_stubs.save`
    """
    return (yield Msg("create", name=name))


@plan
def save() -> MsgGenerator:
    """
    Close a bundle of readings and emit a completed Event document.

    Yields
    ------
    msg : Msg
        Msg('save')

    See Also
    --------
    :func:`bluesky.plan_stubs.create`
    """
    return (yield Msg("save"))


@plan
def drop() -> MsgGenerator:
    """
    Drop a bundle of readings without emitting a completed Event document.

    Yields
    ------
    msg : Msg
        Msg('drop')

    See Also
    --------
    :func:`bluesky.plan_stubs.save`
    :func:`bluesky.plan_stubs.create`
    """
    return (yield Msg("drop"))


@plan
def read(obj: Readable) -> MsgGenerator[Reading]:
    """
    Take a reading and add it to the current bundle of readings.

    Parameters
    ----------
    obj : Device or Signal

    Yields
    ------
    msg : Msg
        Msg('read', obj)

    Returns
    -------
    reading :
        Reading object representing information recorded
    """
    return (yield Msg("read", obj))


@typing.overload
def locate(obj: Locatable, squeeze: Literal[True] = True) -> Location: ...  # type: ignore[overload-overlap]
@typing.overload
def locate(*objs: Locatable, squeeze: bool = True) -> list[Location]: ...
@plan
def locate(*objs, squeeze=True):
    """
    Locate some Movables and return their locations.

    Parameters
    ----------
    obj : Device or Signal
    sqeeze: bool
        If True, return the result as a list.
        If False, always return a list of retults even with a single object.

    Yields
    ------
     msg : Msg
        ``Msg('locate', obj1, ..., objn, squeeze=True)``
    """
    return (yield Msg("locate", *objs, squeeze=squeeze))


@plan
def monitor(obj: Readable, *, name: Optional[str] = None, **kwargs) -> MsgGenerator:
    """
    Asynchronously monitor for new values and emit Event documents.

    Parameters
    ----------
    obj : Device or Signal
    args :
        passed through to ``obj.subscribe()``
    name : string, optional
        name of event stream; default is None
    kwargs :
        passed through to ``obj.subscribe()``

    Yields
    ------
    msg : Msg
        ``Msg('monitor', obj, *args, **kwargs)``

    See Also
    --------
    :func:`bluesky.plan_stubs.unmonitor`
    """
    return (yield Msg("monitor", obj, name=name, **kwargs))


@plan
def unmonitor(obj: Readable) -> MsgGenerator:
    """
    Stop monitoring.

    Parameters
    ----------
    obj : Device or Signal

    Yields
    ------
    msg : Msg
        Msg('unmonitor', obj)

    See Also
    --------
    :func:`bluesky.plan_stubs.monitor`
    """
    return (yield Msg("unmonitor", obj))


@plan
def null() -> MsgGenerator:
    """
    Yield a no-op Message. (Primarily for debugging and testing.)

    Yields
    ------
    msg : Msg
        Msg('null')
    """
    return (yield Msg("null"))


@plan
def abs_set(
    obj: Movable,
    *args: Any,
    group: Optional[Hashable] = None,
    wait: bool = False,
    **kwargs,
) -> MsgGenerator[Status]:
    """
    Set a value. Optionally, wait for it to complete before continuing.

    Parameters
    ----------
    obj : Device
    args :
        passed to obj.set()
    group : string (or any hashable object), optional
        identifier used by 'wait'
    wait : boolean, optional
        If True, wait for completion before processing any more messages.
        False by default.
    kwargs :
        passed to obj.set()

    Yields
    ------
    msg : Msg

    Returns
    -------
    status :
        Status that completes when the value is set. If `wait` is True,
        this will always be complete by the time it is returned.

    See Also
    --------
    :func:`bluesky.plan_stubs.rel_set`
    :func:`bluesky.plan_stubs.wait`
    :func:`bluesky.plan_stubs.mv`
    """
    if wait and group is None:
        group = str(uuid.uuid4())
    ret = yield Msg("set", obj, *args, group=group, **kwargs)
    if wait:
        yield Msg("wait", None, group=group)
    return ret


@plan
def rel_set(
    obj: Movable,
    *args: Any,
    group: Optional[Hashable] = None,
    wait: bool = False,
    **kwargs,
) -> MsgGenerator[Status]:
    """
    Set a value relative to current value. Optionally, wait before continuing.

    Parameters
    ----------
    obj : Device
    args :
        passed to obj.set()
    group : string (or any hashable object), optional
        identifier used by 'wait'; None by default
    wait : boolean, optional
        If True, wait for completion before processing any more messages.
        False by default.
    kwargs :
        passed to obj.set()

    Yields
    ------
    msg : Msg

    Returns
    -------
    status :
        Status that completes when the value is set. If `wait` is True,
        this will always be complete by the time it is returned.

    See Also
    --------
    :func:`bluesky.plan_stubs.abs_set`
    :func:`bluesky.plan_stubs.wait`
    """
    from .preprocessors import relative_set_wrapper

    return (yield from relative_set_wrapper(abs_set(obj, *args, group=group, wait=wait, **kwargs)))


# The format (device1, value1, device2, value2, ...)
# is not currently able to be represented in python's type system
@plan
def mv(
    *args: Union[Movable, Any],
    group: Optional[Hashable] = None,
    timeout: Optional[float] = None,
    **kwargs,
) -> MsgGenerator[tuple[Status, ...]]:
    """
    Move one or more devices to a setpoint. Wait for all to complete.

    If more than one device is specified, the movements are done in parallel.

    Parameters
    ----------
    args :
        device1, value1, device2, value2, ...
    group : string, optional
        Used to mark these as a unit to be waited on.
    timeout : float, optional
        Specify a maximum time that the move(s) can be waited for.
    kwargs :
        passed to obj.set()

    Yields
    ------
    msg : Msg

    Returns
    -------
    statuses :
        Tuple of n statuses, one for each move operation

    See Also
    --------
    :func:`bluesky.plan_stubs.abs_set`
    :func:`bluesky.plan_stubs.mvr`
    """
    group = group or str(uuid.uuid4())
    status_objects = []

    cyl = reduce(operator.add, [cycler(obj, [val]) for obj, val in partition(2, args)])
    (step,) = merge_cycler(cyl)
    for obj, val in step.items():
        ret = yield Msg("set", obj, val, group=group, **kwargs)
        status_objects.append(ret)
    yield Msg("wait", None, group=group, timeout=timeout)
    return tuple(status_objects)


mov = mv  # synonym


@plan
def mvr(
    *args: Union[Movable, Any], group: Optional[Hashable] = None, timeout: Optional[float] = None, **kwargs
) -> MsgGenerator[tuple[Status, ...]]:
    """
    Move one or more devices to a relative setpoint. Wait for all to complete.

    If more than one device is specified, the movements are done in parallel.

    Parameters
    ----------
    args :
        device1, value1, device2, value2, ...
    group : string, optional
        Used to mark these as a unit to be waited on.
    timeout : float, optional
        Specify a maximum time that the move(s) can be waited for.
    kwargs :
        passed to obj.set()

    Yields
    ------
    msg : Msg

    Returns
    -------
    statuses :
        Tuple of n statuses, one for each move operation

    See Also
    --------
    :func:`bluesky.plan_stubs.rel_set`
    :func:`bluesky.plan_stubs.mv`
    """
    objs = []
    for obj, val in partition(2, args):  # noqa: B007
        objs.append(obj)

    from .preprocessors import relative_set_decorator

    @relative_set_decorator(objs)
    def inner_mvr():
        return (yield from mv(*args, group=group, timeout=timeout, **kwargs))

    return (yield from inner_mvr())


movr = mvr  # synonym


@plan
def rd(obj: Readable, *, default_value: Any = 0) -> MsgGenerator[Any]:
    """Reads a single-value non-triggered object

    This is a helper plan to get the scalar value out of a Device
    (such as an EpicsMotor or a single EpicsSignal).

    For devices that implement the Locatable protocol, the location is canonical
    and is returned without parsing the read keys.

    For devices that have more than one read key the following rules are used:

    - if exactly 1 field is hinted that value is used
    - if no fields are hinted and there is exactly 1 value in the
      reading that value is used
    - if more than one field is hinted an Exception is raised
    - if no fields are hinted and there is more than one key in the reading an
      Exception is raised

    The devices is not triggered and this plan does not create any Events

    Parameters
    ----------
    obj : Device
        The device to be read

    default_value : Any
        The value to return when not running in a "live" RunEngine.
        This come ups when ::

           ret = yield Msg('read', obj)
           assert ret is None

        the plan is passed to `list` or some other iterator that
        repeatedly sends `None` into the plan to advance the
        generator.

    Returns
    -------
    val : Any or None
        The "single" value of the device

    """
    # Location is canonical if it exists
    if isinstance(obj, Locatable):
        location = yield Msg("locate", obj)
        if location is None:
            # list-ify mode
            return default_value
        else:
            return location["readback"]

    hints = get_hinted_fields(obj)
    if len(hints) > 1:
        msg = (
            f"Your object {obj} ({obj.name}.{getattr(obj, 'dotted_name', '')}) "
            f"has {len(hints)} items hinted ({hints}).  We do not know how to "
            "pick out a single value.  Please adjust the hinting by setting the "
            "kind of the components of this device or by reading one of its components"
        )
        raise ValueError(msg)
    elif len(hints) == 0:
        hint = None
        if hasattr(obj, "read_attrs"):
            if len(obj.read_attrs) != 1:
                msg = (
                    f"Your object {obj} ({obj.name}.{getattr(obj, 'dotted_name', '')}) "
                    f"and has {len(obj.read_attrs)} read attrs.  We do not know how to "
                    "pick out a single value.  Please adjust the hinting/read_attrs by "
                    "setting the kind of the components of this device or by reading one "
                    "of its components"
                )

                raise ValueError(msg)
    # len(hints) == 1
    else:
        (hint,) = hints

    ret = yield from read(obj)

    # list-ify mode
    if ret is None:
        return default_value

    if hint is not None:
        return ret[hint]["value"]  # type: ignore

    # handle the no hint 1 field case
    try:
        (data,) = ret.values()
    except ValueError as er:
        msg = (
            f"Your object {obj} ({obj.name}.{getattr(obj, 'dotted_name', '')}) "
            f"and has {len(ret)} read values.  We do not know how to pick out a "
            "single value.  Please adjust the hinting/read_attrs by setting the "
            "kind of the components of this device or by reading one of its components"
        )

        raise ValueError(msg) from er
    else:
        return data["value"]  # type: ignore


@plan
def stop(obj: Stoppable) -> MsgGenerator:
    """
    Stop a device.

    Parameters
    ----------
    obj : Device

    Yields
    ------
    msg : Msg
    """
    return (yield Msg("stop", obj))


@plan
def trigger(
    obj: Triggerable,
    *,
    group: Optional[Hashable] = None,
    wait: bool = False,
) -> MsgGenerator[Status]:
    """
    Trigger and acquisition. Optionally, wait for it to complete.

    Parameters
    ----------
    obj : Device
    group : string (or any hashable object), optional
        identifier used by 'wait'; None by default
    wait : boolean, optional
        If True, wait for completion before processing any more messages.
        False by default.

    Yields
    ------
    msg : Msg

    Returns
    -------
    status :
        Status that completes when trigger is complete. If `wait` is True,
        this will always be complete by the time it is returned.

    """
    ret = yield Msg("trigger", obj, group=group)
    if wait:
        yield Msg("wait", None, group=group)
    return ret


@plan
def sleep(time: float) -> MsgGenerator:
    """
    Tell the RunEngine to sleep, while asynchronously doing other processing.

    This is not the same as ``import time; time.sleep()`` because it allows
    other actions, like interruptions, to be processed during the sleep.

    Parameters
    ----------
    time : float
        seconds

    Yields
    ------
    msg : Msg
        Msg('sleep', None, time)
    """
    return (yield Msg("sleep", None, time))


@plan
def wait(
    group: Optional[Hashable] = None,
    *,
    timeout: Optional[float] = None,
    error_on_timeout: bool = True,
    watch: Sequence[str] = (),
):
    """
    Wait for all statuses in a group to report being finished.

    Parameters
    ----------
    group : string (or any hashable object), optional
        Identifier given to `abs_set`, `rel_set`, `trigger`; None by default
    timeout : float, optional
        The maximum duration, in seconds, to wait for all objects in the group to complete.
        If the timeout expires and `error_on_timeout` is set to True, a TimeoutError is raised.

    error_on_timeout : bool, Defaults to True
        Specifies the behavior when the timeout is reached:
        - If True, a TimeoutError is raised if the operations do not complete within the specified timeout.
        - If False, the method returns once all objects are done.
    watch : set of watch groups, optional
        Additional groups to monitor while waiting for the primary group. Raises an exception if any watched group
        fails.
    Yields
    ------
    msg : Msg
        Msg('wait', None, group=group, error_on_timeout=error_on_timeout, timeout=timeout)
    """
    return (yield Msg("wait", None, group=group, error_on_timeout=error_on_timeout, timeout=timeout, watch=watch))


_wait = wait  # for internal references to avoid collision with 'wait' kwarg


@plan
def checkpoint() -> MsgGenerator:
    """
    If interrupted, rewind to this point.

    Yields
    ------
    msg : Msg
        Msg('checkpoint')

    See Also
    --------
    :func:`bluesky.plan_stubs.clear_checkpoint`
    """
    return (yield Msg("checkpoint"))


@plan
def clear_checkpoint() -> MsgGenerator:
    """
    Designate that it is not safe to resume. If interrupted or paused, abort.

    Yields
    ------
    msg : Msg
        Msg('clear_checkpoint')

    See Also
    --------
    :func:`bluesky.plan_stubs.checkpoint`
    """
    return (yield Msg("clear_checkpoint"))


@plan
def pause() -> MsgGenerator:
    """
    Pause and wait for the user to resume.

    Yields
    ------
    msg : Msg
        Msg('pause')

    See Also
    --------
    :func:`bluesky.plan_stubs.deferred_pause`
    :func:`bluesky.plan_stubs.sleep`
    """
    return (yield Msg("pause", None, defer=False))


@plan
def deferred_pause() -> MsgGenerator:
    """
    Pause at the next checkpoint.

    Yields
    ------
    msg : Msg
        Msg('pause', defer=True)

    See Also
    --------
    :func:`bluesky.plan_stubs.pause`
    :func:`bluesky.plan_stubs.sleep`
    """
    return (yield Msg("pause", None, defer=True))


@plan
def input_plan(prompt: str = "") -> MsgGenerator[str]:
    """
    Prompt the user for text input.

    Parameters
    ----------
    prompt : str
        prompt string, e.g., 'enter user name' or 'enter next position'

    Yields
    ------
    msg : Msg
        Msg('input', prompt=prompt)

    Returns
    -------
    input :
    """
    return (yield Msg("input", prompt=prompt))


@plan
def prepare(obj: Preparable, *args, group: Optional[Hashable] = None, wait: bool = False, **kwargs):
    """
    Prepare a device ready for trigger or kickoff.

    Parameters
    ----------
    obj : Preparable
        Device with 'prepare' method
    group : string (or any hashable object), optional
        identifier used by 'wait'
    wait : boolean, optional
        If True, wait for completion before processing any more messages.
        False by default.
    kwargs
        passed through to ``obj.prepare()``

    Yields
    ------
    msg : Msg
        Msg('kickoff', obj)

    See Also
    --------
    :func:`bluesky.plan_stubs.complete`
    :func:`bluesky.plan_stubs.collect`
    :func:`bluesky.plan_stubs.wait`
    """
    ret = yield Msg("prepare", obj, *args, group=group, **kwargs)
    if wait:
        yield from _wait(group=group)
    return ret


@plan
def kickoff(
    obj: Flyable,
    *,
    group: Optional[Hashable] = None,
    wait: bool = False,
    **kwargs,
) -> MsgGenerator[Status]:
    """
    Kickoff one fly-scanning device.

    Parameters
    ----------
    obj : fly-able Device with 'kickoff', and 'complete' methods.
    group : string (or any hashable object), optional
        identifier used by 'wait'.
    wait : boolean, optional
        If True, wait for completion before processing any more messages.
        False by default.
    kwargs
        passed through to ``obj.kickoff()``

    Yields
    ------
    msg : Msg
        Msg('kickoff', obj)

    Returns
    -------
    status :
        Status of kickoff operation. If `wait` is True,
        this will always be complete by the time it is returned.

    See Also
    --------
    :func:`bluesky.plan_stubs.complete`
    :func:`bluesky.plan_stubs.collect`
    :func:`bluesky.plan_stubs.wait`
    """
    ret = yield Msg("kickoff", obj, group=group, **kwargs)
    if wait:
        yield from _wait(group=group)
    return ret


@plan
def kickoff_all(*args, group: Optional[Hashable] = None, wait: bool = True, **kwargs):
    """
    Kickoff one or more fly-scanning devices.

    Parameters
    ----------
    *args : Any fly-able
        Device with 'kickoff', and 'complete' methods.
    group : string (or any hashable object), optional
        identifier used by 'wait'.
    wait : boolean, optional
        If True, wait for completion before processing any more messages.
        True by default.
    kwargs
        passed through to 'kickoff' for each device

    Yields
    ------
    msg : Msg
        Msg('kickoff', obj)

    See Also
    --------
    :func:`bluesky.plan_stubs.complete`
    :func:`bluesky.plan_stubs.collect`
    :func:`bluesky.plan_stubs.wait`
    """
    objs = [check_supports(arg, Flyable) for arg in args]
    group = group or str(uuid.uuid4())
    statuses: list[Status] = []

    for obj in objs:
        ret = yield Msg("kickoff", obj, group=group, **kwargs)
        statuses.append(ret)
    if wait:
        yield from _wait(group=group)

    return tuple(statuses)


@plan
def complete(
    obj: Flyable,
    *,
    group: Optional[Hashable] = None,
    wait: bool = False,
    **kwargs,
) -> MsgGenerator[Status]:
    """
    Tell a flyable, 'stop collecting, whenever you are ready'.

    A flyable returns a status object. Some flyers respond to this
    command by stopping collection and returning a finished status
    object immediately. Other flyers finish their given course and
    finish whenever they finish, irrespective of when this command is
    issued.

    Parameters
    ----------
    obj : fly-able
        Device with 'kickoff' and 'complete' methods.
    group : string (or any hashable object), optional
        identifier used by 'wait'
    wait : boolean, optional
        If True, wait for completion before processing any more messages.
        False by default.
    kwargs
        passed through to ``obj.complete()``

    Yields
    ------
    msg : Msg
        a 'complete' Msg and maybe a 'wait' message

    Returns
    -------
    status :
        Status of complete operation. If `wait` is True,
        this will always be complete by the time it is returned.

    See Also
    --------
    :func:`bluesky.plan_stubs.kickoff`
    :func:`bluesky.plan_stubs.collect`
    :func:`bluesky.plan_stubs.wait`
    """
    ret = yield Msg("complete", obj, group=group, **kwargs)
    if wait:
        yield from _wait(group=group)
    return ret


@plan
def complete_all(*args, group: Optional[Hashable] = None, wait: bool = False, **kwargs):
    """
    Tell one or more flyable objects, 'stop collecting, whenever you are ready'.

    A flyable returns a status object. Some flyers respond to this
    command by stopping collection and returning a finished status
    object immediately. Other flyers finish their given course and
    finish whenever they finish, irrespective of when this command is
    issued.

    Parameters
    ----------
    *args : Any fly-able
        Device with 'kickoff' and 'complete' methods.
    group : string (or any hashable object), optional
        identifier used by 'wait'
    wait : boolean, optional
        If True, wait for completion before processing any more messages.
        False by default.
    kwargs
        passed through to 'complete' for each device

    Yields
    ------
    msg : Msg
        a 'complete' Msg and maybe a 'wait' message

    See Also
    --------
    :func:`bluesky.plan_stubs.kickoff`
    :func:`bluesky.plan_stubs.collect`
    :func:`bluesky.plan_stubs.wait`
    """
    objs = [check_supports(arg, Flyable) for arg in args]
    group = group or str(uuid.uuid4())
    statuses: list[Status] = []

    for obj in objs:
        ret = yield Msg("complete", obj, group=group, **kwargs)
        statuses.append(ret)
    if wait:
        yield from _wait(group=group)

    return tuple(statuses)


@plan
def collect(
    obj: Flyable, *args, stream: bool = False, return_payload: bool = True, name: Optional[str] = None
) -> MsgGenerator[list[PartialEvent]]:
    """
    Collect data cached by one or more fly-scanning devices and emit documents.

    Parameters
    ----------
    obj : A device with 'kickoff', 'complete', and 'collect' methods.
    stream : boolean, optional
        If False (default), emit Event documents in one bulk dump. If True,
        emit events one at time.
    return_payload: boolean, optional
        If True (default), return the collected Events. If False, return None.
        Using ``stream=True`` and ``return_payload=False`` together avoids
        accumulating the documents in memory: they are emitted as they are
        collected, and they are not accumulated.
    name: str, optional
        If not None, will collect for the named string specifically, else collect will be performed
        on all streams.

    Yields
    ------
    msg : Msg
        Msg('collect', obj)

    See Also
    --------
    :func:`bluesky.plan_stubs.kickoff`
    :func:`bluesky.plan_stubs.complete`
    :func:`bluesky.plan_stubs.wait`
    """
    return (yield Msg("collect", obj, *args, stream=stream, return_payload=return_payload, name=name))


@plan
def collect_while_completing(flyers, dets, flush_period=None, stream_name=None, watch: Sequence[str] = ()):
    """
    Collect data from one or more fly-scanning devices and emit documents, then collect and emit
    data from one or more Collectable detectors until all are done.

    Parameters
    ----------
    flyers: An iterable sequence of fly-able devices with 'kickoff', 'complete' and
        'collect' methods.
    dets: An iterable sequence of collectable devices with 'describe_collect' method.
    flush_period: float, int
        Time period in seconds between each yield from collect while waiting for triggered
        objects to be done
    stream_name: str, optional
        If not None, will collect for the named string specifically, else collect will be performed
        on all streams.
    watch: set of watch groups, optional
        Additional groups to monitor while collecting from flyers.
    Yields
    ------
    msg : Msg
        A 'complete' message or 'collect' message

    See Also
    --------
    :func:`bluesky.plan_stubs.complete`
    :func:`bluesky.plan_stubs.collect`
    """
    group = short_uid(label="complete")
    yield from complete_all(*flyers, group=group, wait=False)
    done = False
    while not done:
        done = yield from wait(group=group, timeout=flush_period, error_on_timeout=False, watch=watch)
        yield from collect(*dets, name=stream_name)


@plan
def configure(
    obj: Configurable,
    *args,
    **kwargs,
) -> MsgGenerator[Mapping[str, Reading]]:
    """
    Change Device configuration and emit an updated Event Descriptor document.

    Parameters
    ----------
    obj : Device
    args
        passed through to ``obj.configure()``
    kwargs
        passed through to ``obj.configure()``

    Yields
    ------
    msg : Msg
        ``Msg('configure', obj, *args, **kwargs)``

    Returns
    -------
    configuration:
        Tuple of old and new configuration as returned by
        obj.read_configuration()
    """
    return (yield Msg("configure", obj, *args, **kwargs))


@plan
def stage(
    obj: Stageable,
    *,
    group: Optional[Hashable] = None,
    wait: Optional[bool] = None,
) -> MsgGenerator[Union[Status, list[Any]]]:
    """
    'Stage' a device (i.e., prepare it for use, 'arm' it).

    Parameters
    ----------
    obj : Device
    group : string (or any hashable object), optional
        identifier used by 'wait'; None by default
    wait : boolean, optional
        If True, wait for completion before processing any more messages.
        False by default.

    Yields
    ------
    msg : Msg

    Returns
    -------
    stage :
        Either a status representing the stage operation or a list of
        staged values for backward compatibility.

    See Also
    --------
    :func:`bluesky.plan_stubs.unstage`
    :func:`bluesky.plan_stubs.stage_all`
    """
    ret = yield Msg("stage", obj, group=group)
    old_style = not isinstance(ret, Status)
    if old_style:
        if (wait is None) or wait:
            # Old-style devices will just block. We do not need to explicitly wait.
            pass
        else:  # wait is False-y
            # No way to tell old-style devices not to wait
            raise RuntimeError(f"{obj}: Is an old style device and cannot be told not to wait")
    else:
        if wait:
            yield Msg("wait", None, group=group)
    return ret


@plan
def stage_all(
    *args: Stageable,
    group: Optional[Hashable] = None,
) -> MsgGenerator[None]:
    """
    'Stage' one or more devices (i.e., prepare them for use, 'arm' them).

    Parameters
    ----------
    args :
        device1, device2, device3, ...
    group : string (or any hashable object), optional
        identifier used by 'wait'; None by default

    Yields
    ------
    msg : Msg

    See Also
    --------
    :func:`bluesky.plan_stubs.stage`
    :func:`bluesky.plan_stubs.unstage_all`
    """
    group = group or str(uuid.uuid4())
    status_objects = []

    for obj in args:
        ret = yield Msg("stage", obj, group=group)
        if isinstance(ret, Status):
            status_objects.append(ret)

    if status_objects:
        yield Msg("wait", None, group=group)


@plan
def unstage(
    obj: Stageable,
    *,
    group: Optional[Hashable] = None,
    wait: Optional[bool] = None,
) -> MsgGenerator[Union[Status, list[Any]]]:
    """
    'Unstage' a device (i.e., put it in standby, 'disarm' it).

    Parameters
    ----------
    obj : Device
    group : string (or any hashable object), optional
        identifier used by 'wait'; None by default
    wait : boolean, optional
        If True, wait for completion before processing any more messages.
        False by default.

    Yields
    ------
    msg : Msg

    Returns
    -------
    unstage :
        Either a status representing the stage operation or a list of
        staged values for backward compatibility.

    See Also
    --------
    :func:`bluesky.plan_stubs.stage`
    :func:`bluesky.plan_stubs.unstage_all`
    """
    ret = yield Msg("unstage", obj, group=group)
    old_style = not isinstance(ret, Status)
    if old_style:
        if (wait is None) or wait:
            # Old-style devices will just block. We do not need to explicitly wait.
            pass
        else:
            # No way to tell old-style devices not to wait
            raise RuntimeError(f"{obj}: Is an old style device and cannot be told not to wait")
    else:
        if wait:
            yield Msg("wait", None, group=group)
    return ret


@plan
def unstage_all(*args: Stageable, group: Optional[Hashable] = None) -> MsgGenerator[None]:
    """
    'Unstage' one or more devices (i.e., put them in standby, 'disarm' them).

    Parameters
    ----------
    args :
        device1, device2, device3, ...
    group : string (or any hashable object), optional
        identifier used by 'wait'; None by default

    Yields
    ------
    msg : Msg

    See Also
    --------
    :func:`bluesky.plan_stubs.unstage`
    :func:`bluesky.plan_stubs.stage_all`
    """
    group = group or str(uuid.uuid4())
    status_objects = []

    for obj in args:
        ret = yield Msg("unstage", obj, group=group)
        if isinstance(ret, Status):
            status_objects.append(ret)

    if status_objects:
        yield Msg("wait", None, group=group)


@plan
def subscribe(name: str, func: Callable[[str, Mapping[str, Any]], None]) -> MsgGenerator[int]:
    """
    Subscribe the stream of emitted documents.

    Parameters
    ----------
    name : {'all', 'start', 'descriptor', 'event', 'stop'}
    func : callable
        Expected signature: ``f(name, doc)`` where ``name`` is one of the
        strings above ('all, 'start', ...) and ``doc`` is a dict

    Yields
    ------
    msg : Msg
        Msg('subscribe', None, func, name)

    Returns
    -------
    token :
        Unique identifier for a subscription

    See Also
    --------
    :func:`bluesky.plan_stubs.unsubscribe`
    """
    return (yield Msg("subscribe", None, func, name))


@plan
def unsubscribe(token: int) -> MsgGenerator:
    """
    Remove a subscription.

    Parameters
    ----------
    token : int
        token returned by processing a 'subscribe' message

    Yields
    ------
    msg : Msg
        Msg('unsubscribe', token=token)

    See Also
    --------
    :func:`bluesky.plan_stubs.subscribe`
    """
    return (yield Msg("unsubscribe", token=token))


@plan
def install_suspender(suspender: SuspenderBase) -> MsgGenerator:
    """
    Install a suspender during a plan.

    Parameters
    ----------
    suspender : :class:`bluesky.suspenders.SuspenderBase`
        The suspender to install

    Yields
    ------
    msg : Msg
        Msg('install_suspender', None, suspender)

    See Also
    --------
    :func:`bluesky.plan_stubs.remove_suspender`
    """
    return (yield Msg("install_suspender", None, suspender))


@plan
def remove_suspender(suspender: SuspenderBase) -> MsgGenerator:
    """
    Remove a suspender during a plan.

    Parameters
    ----------
    suspender : :class:`bluesky.suspenders.SuspenderBase`
        The suspender to remove

    Yields
    ------
    msg : Msg
        Msg('remove_suspender', None, suspender)

    See Also
    --------
    :func:`bluesky.plan_stubs.install_suspender`
    """
    return (yield Msg("remove_suspender", None, suspender))


@plan
def open_run(md: Optional[CustomPlanMetadata] = None) -> MsgGenerator[str]:
    """
    Mark the beginning of a new 'run'. Emit a RunStart document.

    Parameters
    ----------
    md : dict, optional
        metadata

    Yields
    ------
    msg : Msg
        ``Msg('open_run', **md)``

    Returns
    -------
    uuid :
        Unique ID for the run

    See Also
    --------
    :func:`bluesky.plans_stubs.close_run`
    """
    return (yield Msg("open_run", **(md or {})))


@plan
def close_run(exit_status: Optional[str] = None, reason: Optional[str] = None) -> MsgGenerator[str]:
    """
    Mark the end of the current 'run'. Emit a RunStop document.

    Parameters
    ----------
    exit_status : {None, 'success', 'abort', 'fail'}
        The exit status to report in the Stop document
    reason : str, optional
        Long-form description of why the run ended

    Yields
    ------
    msg : Msg
        Msg('close_run')

    Returns
    -------
    uuid :
        Unique ID for the run

    See Also
    --------
    :func:`bluesky.plans_stubs.open_run`
    """
    return (yield Msg("close_run", exit_status=exit_status, reason=reason))


@plan
def wait_for(futures: Iterable[Callable[[], Awaitable[Any]]], **kwargs) -> MsgGenerator:
    """
    Low-level: wait for a list of ``asyncio.Future`` objects to set (complete).

    Parameters
    ----------
    futures : iterable
        iterable collection of coroutine functions that take no arguments
    kwargs
        passed through to ``asyncio.wait()``

    Yields
    ------
    msg : Msg
        ``Msg('wait_for', None, futures, **kwargs)``

    See Also
    --------
    :func:`bluesky.plan_stubs.wait`
    """
    return (yield Msg("wait_for", None, futures, **kwargs))


@plan
def trigger_and_read(devices: Sequence[Readable], name: str = "primary") -> MsgGenerator[Mapping[str, Reading]]:
    """
    Trigger and read a list of detectors and bundle readings into one Event.

    Parameters
    ----------
    devices : list
        devices to trigger (if they have a trigger method) and then read
    name : string, optional
        event stream name, a convenient human-friendly identifier; default
        name is 'primary'

    Returns
    -------
    readings:
        dict of device name to recorded information

    Yields
    ------
    msg : Msg
        messages to 'trigger', 'wait' and 'read'
    """
    from .preprocessors import contingency_wrapper

    # If devices is empty, don't emit 'create'/'save' messages.
    if not devices:
        yield from null()
    devices = separate_devices(devices)  # remove redundant entries
    rewindable = all_safe_rewind(devices)  # if devices can be re-triggered

    def inner_trigger_and_read():
        grp = _short_uid("trigger")
        no_wait = True
        for obj in devices:
            if isinstance(obj, Triggerable):
                no_wait = False
                yield from trigger(obj, group=grp)
        # Skip 'wait' if none of the devices implemented a trigger method.
        if not no_wait:
            yield from wait(group=grp)
        yield from create(name)

        def read_plan():
            ret = {}  # collect and return readings to give plan access to them
            for obj in devices:
                reading = yield from read(obj)
                if reading is not None:
                    ret.update(reading)
            return ret

        def standard_path():
            yield from save()

        def exception_path(exp):
            yield from drop()
            raise exp

        ret = yield from contingency_wrapper(read_plan(), except_plan=exception_path, else_plan=standard_path)
        return ret

    from .preprocessors import rewindable_wrapper

    return (yield from rewindable_wrapper(inner_trigger_and_read(), rewindable))


@plan
def broadcast_msg(
    command: str,
    objs: Iterable[Any],
    *args,
    **kwargs,
) -> MsgGenerator[Any]:
    """
    Generate many copies of a message, applying it to a list of devices.

    Parameters
    ----------
    command : string
    objs : iterable
    ``*args``
        args for message
    ``**kwargs``
        kwargs for message

    Yields
    ------
    msg : Msg

    Returns
    -------
    any : out from RunEngine, if any
    """
    return_vals = []
    for o in objs:
        ret = yield Msg(command, o, *args, **kwargs)
        return_vals.append(ret)

    return return_vals


@plan
def repeater(
    n: Optional[int],
    gen_func: Callable[..., MsgGenerator],
    *args,
    **kwargs,
) -> MsgGenerator[None]:
    """
    Generate n chained copies of the messages from gen_func

    Parameters
    ----------
    n : int or None
        total number of repetitions; if None, infinite
    gen_func : callable
        returns generator instance
    ``*args``
        args for gen_func
    ``**kwargs``
        kwargs for gen_func

    Yields
    ------
    msg : Msg

    See Also
    --------
    :func:`bluesky.plan_stubs.caching_repeater`
    """
    it: Any
    it = range
    if n is None:
        n = 0
        it = itertools.count

    for j in it(n):  # noqa: B007
        yield from gen_func(*args, **kwargs)


@plan
def caching_repeater(n: Optional[int], plan: MsgGenerator) -> MsgGenerator[None]:
    """
    Generate n chained copies of the messages in a plan.

    This is different from ``repeater`` above because it takes in a
    generator or iterator, not a function that returns one.

    Parameters
    ----------
    n : int or None
        total number of repetitions; if None, infinite
    plan : iterable

    Yields
    ------
    msg : Msg

    See Also
    --------
    :func:`bluesky.plan_stubs.repeater`
    """
    warnings.warn("The caching_repeater will be removed in a future version of bluesky.", stacklevel=2)
    gen: Any
    if n is None:
        gen = itertools.count(0)
    else:
        gen = range(n)

    lst_plan = list(plan)
    for _ in gen:
        yield from (m for m in lst_plan)


@plan
def one_shot(detectors: Sequence[Readable], take_reading: Optional[TakeReading] = None) -> MsgGenerator[None]:
    """Inner loop of a count.

    This is the default function for ``per_shot`` in count plans.

    Parameters
    ----------
    detectors : Sequence[OphydObj]
        devices to read

    take_reading : plan, optional
        function to do the actual acquisition ::

           def take_reading(dets, name='primary'):
                yield from ...

        Callable[List[OphydObj], Optional[str]] -> Generator[Msg], optional

        Defaults to `trigger_and_read`

    Yields
    ------
    msg : Msg
    """
    take_reading = trigger_and_read if take_reading is None else take_reading
    yield Msg("checkpoint")
    yield from take_reading(list(detectors))  # type: ignore  # Movable issue


@plan
def one_1d_step(
    detectors: Sequence[Readable],
    motor: Movable,
    step: Any,
    take_reading: Optional[TakeReading] = None,
) -> MsgGenerator[Mapping[str, Reading]]:
    """
    Inner loop of a 1D step scan

    This is the default function for ``per_step`` param in 1D plans.

    Parameters
    ----------
    detectors : list or tuple
        devices to read
    motor : Settable
        The motor to move
    step : Any
        Where to move the motor to
    take_reading : plan, optional
        function to do the actual acquisition ::

           def take_reading(dets, name='primary'):
                yield from ...

        Callable[List[OphydObj], Optional[str]] -> Generator[Msg], optional

        Defaults to `trigger_and_read`

    Yields
    ------
    msg : Msg

    Returns
    -------
    readings :
        dict of device names to recorded information
    """
    take_reading = trigger_and_read if take_reading is None else take_reading

    def move():
        grp = _short_uid("set")
        yield Msg("checkpoint")
        yield Msg("set", motor, step, group=grp)
        yield Msg("wait", None, group=grp)

    yield from move()
    return (yield from take_reading(list(detectors) + [motor]))  # type: ignore


@plan
def move_per_step(step: Mapping[Movable, Any], pos_cache: dict[Movable, Any]) -> MsgGenerator[None]:
    """
    Inner loop of an N-dimensional step scan without any readings

    This can be used as a building block for custom ``per_step`` stubs.

    Parameters
    ----------
    step : dict
        mapping motors to positions in this step
    pos_cache : dict
        mapping motors to their last-set positions

    Yields
    ------
    msg : Msg
    """
    yield Msg("checkpoint")
    grp = _short_uid("set")
    for motor, pos in step.items():
        if pos == pos_cache[motor]:
            # This step does not move this motor.
            continue
        yield Msg("set", motor, pos, group=grp)
        pos_cache[motor] = pos
    yield Msg("wait", None, group=grp)


@plan
def one_nd_step(
    detectors: Sequence[Readable],
    step: Mapping[Movable, Any],
    pos_cache: dict[Movable, Any],
    take_reading: Optional[TakeReading] = None,
) -> MsgGenerator[None]:
    """
    Inner loop of an N-dimensional step scan

    This is the default function for ``per_step`` param`` in ND plans.

    Parameters
    ----------
    detectors : list or tuple
        devices to read
    step : dict
        mapping motors to positions in this step
    pos_cache : dict
        mapping motors to their last-set positions
    take_reading : plan, optional
        function to do the actual acquisition ::

           def take_reading(dets, name='primary'):
                yield from ...

        Callable[List[OphydObj], Optional[str]] -> Generator[Msg], optional

        Defaults to `trigger_and_read`

    Yields
    ------
    msg : Msg
    """
    take_reading = trigger_and_read if take_reading is None else take_reading
    motors = step.keys()
    yield from move_per_step(step, pos_cache)
    yield from take_reading(list(detectors) + list(motors))  # type: ignore  # Movable issue


@plan
def repeat(
    plan: Callable[[], MsgGenerator],
    num: Optional[int] = 1,
    delay: ScalarOrIterableFloat = 0.0,
) -> MsgGenerator[Any]:
    """
    Repeat a plan num times with delay and checkpoint between each repeat.

    This is different from ``repeater`` and ``caching_repeater`` in that it
    adds ``checkpoint`` and optionally ``sleep`` messages if delay is provided.
    This is intended for users who need the structure of ``count`` but do not
    want to reimplement the control flow.

    Parameters
    ----------
    plan: callable
        Callable that returns an iterable of Msg objects
    num : integer, optional
        number of readings to take; default is 1

        If None, capture data until canceled
    delay : iterable or scalar, optional
        time delay between successive readings; default is 0

    Yields
    ------
    msg : Msg

    Returns
    -------
    any : output of original plan

    Notes
    -----
    If ``delay`` is an iterable, it must have at least ``num - 1`` entries or
    the plan will raise a ``ValueError`` during iteration.
    """
    # Create finite or infinite counter
    iterator: Iterable
    if num is None:
        iterator = itertools.count()
    else:
        iterator = range(num)

    # If delay is a scalar, repeat it forever. If it is an iterable, leave it.
    if not isinstance(delay, Iterable):
        delay = itertools.repeat(delay)
    else:
        try:
            num_delays = len(delay)  # type: ignore
        except TypeError:
            # No way to tell in advance if we have enough delays.
            pass
        else:
            if num and num - 1 > num_delays:
                raise ValueError("num=%r but delays only provides %r entries" % (num, num_delays))  # noqa: UP031
        delay = iter(delay)

    def repeated_plan():
        for i in iterator:
            now = time.time()  # Intercept the flow in its earliest moment.
            yield Msg("checkpoint")
            yield from ensure_generator(plan())
            try:
                d = next(delay)
            except StopIteration as stop:
                if i + 1 == num:
                    break
                elif num is None:
                    break
                else:
                    # num specifies a number of iterations less than delay
                    raise ValueError("num=%r but delays only provides %r entries" % (num, i)) from stop  # noqa: UP031
            if d is not None:
                d = d - (time.time() - now)
                if d > 0:  # Sleep if and only if time is left to do it.
                    yield Msg("sleep", None, d)

    return (yield from repeated_plan())



import time as ttime
import warnings

import numpy as np

from .utils import Msg

try:
    import ophyd.sim  # noqa: F401
except ImportError:
    raise ImportError("""
The simulated hardware objects in the bluesky.examples module have been
moved to ophyd.sim. A version of ophyd v0.8.0 or greater is required.""")  # noqa: B904
else:
    warnings.warn(
        """
The simulated hardware objects in the bluesky.examples module have been
moved to ophyd.sim. Update imports to suppress this warning.
    """,
        stacklevel=2,
    )
    from ophyd.sim import *  # noqa: F403, F401


# These are very old, raw implementations of 'plans' the pre-date the
# bluesky.plans module. They are not documented or (we hope) used in
# production, but they are still used in tests.


def simple_scan(motor):
    yield Msg("open_run")
    yield Msg("set", motor, 5)
    yield Msg("read", motor)
    yield Msg("close_run")


def conditional_break(det, motor, threshold):
    """Set, trigger, read until the detector reads intensity < threshold"""
    i = 0
    yield Msg("open_run")
    while True:
        print("LOOP %d" % i)
        yield Msg("set", motor, i)
        yield Msg("trigger", det)
        reading = yield Msg("read", det)
        if reading[det.name]["value"] < threshold:
            print("DONE")
            yield Msg("close_run")
            break
        i += 1


def sleepy(det, motor):
    "Set, trigger motor, sleep for a fixed time, trigger detector, read"
    yield Msg("open_run")
    yield Msg("set", motor, 5)
    yield Msg("sleep", None, 2)  # units: seconds
    yield Msg("trigger", det)
    yield Msg("read", det)
    yield Msg("close_run")


def do_nothing(timeout=5):
    "Generate 'checkpoint' messages until timeout."
    t = ttime.time()
    yield Msg("open_run")
    while True:
        if ttime.time() > t + timeout:
            break
        ttime.sleep(0.1)
        yield Msg("checkpoint")
    yield Msg("close_run")


def checkpoint_forever():
    # simplest pauseable scan
    yield Msg("open_run")
    while True:
        ttime.sleep(0.1)
        yield Msg("checkpoint")
    yield Msg("close_run")


def wait_one(det, motor):
    "Set, trigger, read"
    yield Msg("open_run")
    yield Msg("set", motor, 5, group="A")  # Add to group 'A'.
    yield Msg("wait", None, group="A")  # Wait for everything in group 'A' to finish.
    yield Msg("trigger", det)
    yield Msg("read", det)
    yield Msg("close_run")


def wait_multiple(det, motors):
    "Set motors, trigger all motors, wait for all motors to move."
    yield Msg("open_run")
    for motor in motors:
        yield Msg("set", motor, 5, group="A")
    # Wait for everything in group 'A' to report done.
    yield Msg("wait", None, group="A")
    yield Msg("trigger", det)
    yield Msg("read", det)
    yield Msg("close_run")


def wait_complex(det, motors):
    "Set motors, trigger motors, wait for all motors to move in groups."
    # Same as above...
    yield Msg("open_run")
    for motor in motors[:-1]:
        yield Msg("set", motor, 5, group="A")

    # ...but put the last motor is separate group.
    yield Msg("set", motors[-1], 5, group="B")
    # Wait for everything in group 'A' to report done.
    yield Msg("wait", None, group="A")
    yield Msg("trigger", det)
    yield Msg("read", det)

    # Wait for everything in group 'B' to report done.
    yield Msg("wait", None, group="B")
    yield Msg("trigger", det)
    yield Msg("read", det)
    yield Msg("close_run")


def conditional_pause(det, motor, defer, include_checkpoint):
    yield Msg("open_run")
    for i in range(5):
        if include_checkpoint:
            yield Msg("checkpoint")
        yield Msg("set", motor, i)
        yield Msg("trigger", det)
        reading = yield Msg("read", det)
        if reading["det"]["value"] < 0.2:
            yield Msg("pause", defer=defer)
        print("I'm not pausing yet.")
    yield Msg("close_run")


def simple_scan_saving(det, motor):
    "Set, trigger, read"
    yield Msg("open_run")
    yield Msg("declare_stream", None, motor, det, name="primary")
    yield Msg("create", name="primary")
    yield Msg("set", motor, 5)
    yield Msg("read", motor)
    yield Msg("trigger", det)
    yield Msg("read", det)
    yield Msg("save")
    yield Msg("close_run")


def stepscan(det, motor):
    yield Msg("open_run")
    yield Msg("declare_stream", None, motor, det, name="primary")
    for i in range(-5, 5):
        yield Msg("create", name="primary")
        yield Msg("set", motor, i)
        yield Msg("trigger", det)
        yield Msg("read", motor)
        yield Msg("read", det)
        yield Msg("save")
    yield Msg("close_run")


def cautious_stepscan(det, motor):
    yield Msg("open_run")
    yield Msg("declare_stream", None, motor, det, name="primary")
    for i in range(-5, 5):
        yield Msg("checkpoint")
        yield Msg("create", name="primary")
        yield Msg("set", motor, i)
        yield Msg("trigger", det)
        ret_m = yield Msg("read", motor)
        ret_d = yield Msg("read", det)
        yield Msg("save")
        print("Value at {m} is {d}. Pausing.".format(m=ret_m[motor.name]["value"], d=ret_d[det.name]["value"]))
        yield Msg("pause", None, defer=True)
    yield Msg("close_run")


def fly_gen(flyer):
    yield Msg("open_run")
    yield Msg("kickoff", flyer, group="fly-kickoff")
    yield Msg("wait", None, group="fly-kickoff")
    yield Msg("complete", flyer, group="fly-complete")
    yield Msg("wait", None, group="fly-complete")
    yield Msg("collect", flyer)
    yield Msg("kickoff", flyer, group="fly-kickoff2")
    yield Msg("wait", None, group="fly-kickoff2")
    yield Msg("complete", flyer, group="fly-complete2")
    yield Msg("wait", None, group="fly-complete2")
    yield Msg("collect", flyer)
    yield Msg("close_run")


def multi_sample_temperature_ramp(
    detector, sample_names, sample_positions, scan_motor, start, stop, step, temp_controller, tstart, tstop, tstep
):
    def read_and_store_temp():
        yield Msg("create", name="primary")
        yield Msg("read", temp_controller)
        yield Msg("save")

    peak_centers = [-1 + 3 * n for n in range(len(sample_names))]
    detector.noise = True

    for idx, temp in enumerate(np.arange(tstart, tstop, tstep)):
        # todo would be cute to have the temperature reduce peak noise
        yield Msg("set", temp_controller, temp)
        for sample, sample_position, peak_pos in zip(sample_names, sample_positions, peak_centers):  # noqa: B007
            yield Msg("open_run", sample_name=sample, target_temp=temp)
            detector.center = peak_pos
            detector.sigma = 0.5 + 0.25 * idx
            detector.noise_factor = 0.05 + idx * 0.1
            yield Msg("declare_stream", None, scan_motor, detector, temp_controller, name="primary")
            for scan_pos in np.arange(start, stop, step):
                yield Msg("set", scan_motor, scan_pos)
                # be super paranoid about the temperature. Grab it before and
                # after each trigger!
                # Capturing the temperature data before and after each
                # trigger is resulting in unintended behavior. Uncomment the
                # two `yield from` statements and run this example to see
                # what I'm talking about.
                # yield from read_and_store_temp()
                yield Msg("trigger", detector)
                # yield from read_and_store_temp()
                yield Msg("create", name="primary")
                yield Msg("read", scan_motor)
                yield Msg("read", detector)
                yield Msg("read", temp_controller)
                # yield Msg('sleep', None, .1)
                yield Msg("save")
            # generate the end of the run document
            yield Msg("close_run")



import pytest

from bluesky.utils import register_transform


@pytest.fixture
def transform_cell():
    IPython = pytest.importorskip("IPython")
    ip = IPython.core.interactiveshell.InteractiveShell()
    register_transform("RE", prefix="<", ip=ip)
    if IPython.__version__ >= "7":
        return ip.transform_cell
    else:
        return ip.input_splitter.transform_cell


def test_register_transform_smoke(transform_cell):
    assert True


@pytest.mark.parametrize("cell", ["a < b\n", "RE(plan(a < b))\n", "for j in range(5):\n    < a\n    <b\n"])
def test_no_transform(transform_cell, cell):
    new_cell = transform_cell(cell)
    assert cell == new_cell


@pytest.mark.parametrize("cell", ["< b\n", "<a\n", '<fly(a, b, c), md={"a": "aardvark"}\n', "  <b\n", "<a, c=d\n"])
def test_transform(transform_cell, cell):
    new_cell = transform_cell(cell)

    assert f"RE({cell.lstrip('< ').strip()})\n" == new_cell



import numpy as np
import pytest
from ophyd.sim import SynGauss, det, motor

from bluesky.callbacks.fitting import PeakStats
from bluesky.plans import scan


def get_ps(x, y, shift=0.5):
    """peak status calculation from CHX algorithm."""
    pytest.importorskip("scipy")
    from scipy.special import erf

    lmfit = pytest.importorskip("lmfit")
    ps = {}
    x = np.array(x)
    y = np.array(y)

    COM = np.sum(x * y) / np.sum(y)
    ps["com"] = COM

    # from Maksim: assume this is a peak profile:
    def is_positive(num):
        return True if num > 0 else False

    # Normalize values first:
    ym = (y - np.min(y)) / (np.max(y) - np.min(y)) - shift  # roots are at Y=0

    positive = is_positive(ym[0])
    list_of_roots = []
    for i in range(len(y)):
        current_positive = is_positive(ym[i])
        if current_positive != positive:
            list_of_roots.append(x[i - 1] + (x[i] - x[i - 1]) / (abs(ym[i]) + abs(ym[i - 1])) * abs(ym[i - 1]))
            positive = not positive
    if len(list_of_roots) >= 2:
        FWHM = abs(list_of_roots[-1] - list_of_roots[0])
        CEN = list_of_roots[0] + 0.5 * (list_of_roots[1] - list_of_roots[0])

        ps["fwhm"] = FWHM
        ps["cen"] = CEN

    else:  # ok, maybe it's a step function..
        print("no peak...trying step function...")
        ym = ym + shift

        def err_func(x, x0, k=2, A=1, base=0):  # erf fit from Yugang
            return base - A * erf(k * (x - x0))

        mod = lmfit.Model(err_func)
        # estimate starting values:
        x0 = np.mean(x)
        # k=0.1*(np.max(x)-np.min(x))
        pars = mod.make_params(x0=x0, k=2, A=1.0, base=0.0)
        result = mod.fit(ym, pars, x=x)
        CEN = result.best_values["x0"]
        FWHM = result.best_values["k"]
        ps["fwhm"] = FWHM
        ps["cen"] = CEN

    return ps


def test_peak_statistics(RE):
    """peak statistics calculation on simple gaussian function"""
    x = "motor"
    y = "det"
    ps = PeakStats(x, y)
    RE.subscribe(ps)
    RE(scan([det], motor, -5, 5, 100))

    fields = ["x", "y", "min", "max", "com", "cen", "crossings", "fwhm", "lin_bkg"]
    for field in fields:
        assert hasattr(ps, field), f"{field} is not an attribute of ps"

    np.allclose(ps.cen, 0, atol=1e-6)
    np.allclose(ps.com, 0, atol=1e-6)
    fwhm_gauss = 2 * np.sqrt(2 * np.log(2))  # theoretical value with std=1
    assert np.allclose(ps.fwhm, fwhm_gauss, atol=1e-2)


def test_peak_statistics_compare_chx(RE):
    """This test focuses on gaussian function with noise."""
    s = np.random.RandomState(1)
    noisy_det_fix = SynGauss(
        "noisy_det_fix",
        motor,
        "motor",
        center=0,
        Imax=1,
        noise="uniform",
        sigma=1,
        noise_multiplier=0.1,
        random_state=s,
    )
    x = "motor"
    y = "noisy_det_fix"
    ps = PeakStats(x, y)
    RE.subscribe(ps)

    RE(scan([noisy_det_fix], motor, -5, 5, 100))
    ps_chx = get_ps(ps.x_data, ps.y_data)

    assert np.allclose(ps.cen, ps_chx["cen"], atol=1e-6)
    assert np.allclose(ps.com, ps_chx["com"], atol=1e-6)
    assert np.allclose(ps.fwhm, ps_chx["fwhm"], atol=1e-6)


def test_peak_statistics_with_derivatives(RE):
    """peak statistics calculation on simple gaussian function with derivatives"""
    x = "motor"
    y = "det"
    num_points = 100
    ps = PeakStats(x, y, calc_derivative_and_stats=True)
    RE.subscribe(ps)
    RE(scan([det], motor, -5, 5, num_points))

    assert hasattr(ps, "derivative_stats")
    der_fields = ["x", "y", "min", "max", "com", "cen", "crossings", "fwhm", "lin_bkg"]
    for field in der_fields:
        assert hasattr(ps.derivative_stats, field), f"{field} is not an attribute of ps.der"

    assert type(ps.derivative_stats.x) is np.ndarray
    assert type(ps.derivative_stats.y) is np.ndarray
    assert type(ps.derivative_stats.min) is tuple
    assert type(ps.derivative_stats.max) is tuple
    assert type(ps.derivative_stats.com) is np.float64
    assert type(ps.derivative_stats.cen) is np.float64
    assert type(ps.derivative_stats.crossings) is np.ndarray
    if len(ps.derivative_stats.crossings) >= 2:
        assert type(ps.derivative_stats.fwhm) is float  # noqa: E721
    else:
        assert ps.derivative_stats.fwhm is None
    assert len(ps.derivative_stats.x) == num_points - 1
    assert len(ps.derivative_stats.y) == num_points - 1
    assert np.allclose(np.diff(ps.y_data), ps.derivative_stats.y, atol=1e-10)



from collections import defaultdict
from functools import partial

from bluesky.callbacks.broker import post_run, verify_files_saved
from bluesky.examples import stepscan


def test_scan_and_get_data(RE, hw, db):
    RE.subscribe(db.insert)
    rs = RE(stepscan(hw.det, hw.motor), group="foo", beamline_id="testing", config={})
    if RE.call_returns_result:
        uid = rs.run_start_uids[0]
    else:
        uid = rs[0]
    hdr = db[uid]
    list(hdr.events())


def test_post_run(RE, hw, db):
    RE.subscribe(db.insert)
    output = defaultdict(list)

    def do_nothing(doctype, doc):
        output[doctype].append(doc)

    RE(stepscan(hw.det, hw.motor), {"stop": [post_run(do_nothing, db=db)]})
    assert len(output)
    assert len(output["start"]) == 1
    assert len(output["stop"]) == 1
    assert len(output["descriptor"]) == 1
    assert len(output["event"]) == 10


def test_verify_files_saved(RE, hw, db):
    RE.subscribe(db.insert)

    vfs = partial(verify_files_saved, db=db)
    RE(stepscan(hw.det, hw.motor), {"stop": vfs})



#!/usr/bin/env python3
#
# bluesky documentation build configuration file, created by
# sphinx-quickstart on Sun Jun  7 15:25:26 2015.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import sys

import requests

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('.'))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.mathjax",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "IPython.sphinxext.ipython_directive",
    "IPython.sphinxext.ipython_console_highlighting",
    "matplotlib.sphinxext.plot_directive",
    "sphinx.ext.intersphinx",
    "numpydoc",
    "sphinxcontrib.mermaid",
]

# Configuration options for plot_directive. See:
# https://github.com/matplotlib/matplotlib/blob/f3ed922d935751e08494e5fb5311d3050a3b637b/lib/matplotlib/sphinxext/plot_directive.py#L81
plot_html_show_source_link = False
plot_html_show_formats = False
plot_pre_code = (
    "from matplotlib import pyplot as plt\n"
    "from bluesky import RunEngine\n"
    "from bluesky.callbacks.best_effort import BestEffortCallback\n"
    "RE = RunEngine({})\n"
    "bec = BestEffortCallback()\n"
    "RE.subscribe(bec)\n"
)

# Generate the API documentation when building
autosummary_generate = True
numpydoc_show_class_members = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "bluesky"
copyright = "2015, Brookhaven National Lab"
author = "Brookhaven National Lab"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
import bluesky  # noqa: E402

version = bluesky.__version__
# The full version, including alpha/beta/rc tags.
release = bluesky.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["src/bluesky/_vendor/"]

# The reST default role (used for this markup: `text`) to use for all
# documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"
github_repo = project
github_user = "bluesky"
switcher_json = f"https://blueskyproject.io/{github_repo}/switcher.json"
switcher_exists = requests.get(switcher_json).ok
if not switcher_exists:
    print(
        "*** Can't read version switcher, is GitHub pages enabled? \n"
        "    Once Docs CI job has successfully run once, set the "
        "Github pages source branch to be 'gh-pages' at:\n"
        f"    https://github.com/{github_user}/{github_repo}/settings/pages",
        file=sys.stderr,
    )

# Theme options for pydata_sphinx_theme
# We don't check switcher because there are 3 possible states for a repo:
# 1. New project, docs are not published so there is no switcher
# 2. Existing project with latest copier template, switcher exists and works
# 3. Existing project with old copier template that makes broken switcher,
#    switcher exists but is broken
# Point 3 makes checking switcher difficult, because the updated copier template
# will fix the switcher at the end of the docs workflow, but never gets a chance
# to complete as the docs build warns and fails.
html_theme_options = {
    "logo": {
        "text": project,
    },
    "use_edit_page_button": True,
    "github_url": f"https://github.com/{github_user}/{github_repo}",
    "icon_links": [
        {
            "name": "PyPI",
            "url": f"https://pypi.org/project/{project}",
            "icon": "fas fa-cube",
        }
    ],
    "switcher": {
        "json_url": switcher_json,
        "version_match": version,
    },
    "check_switcher": False,
    "navbar_end": ["theme-switcher", "icon-links", "version-switcher"],
    # Ensure the toctrees created by autosummary directives are not visible in
    # left-panel navigation
    "navigation_depth": 1,
}

# # A dictionary of values to pass into the template engine’s context for all pages
html_context = {
    "github_user": github_user,
    "github_repo": github_repo,
    "github_version": version,
    "doc_path": "docs",
}

# # If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = False

# # If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = False

# # Logo
html_logo = "images/bluesky-logo-dark.svg"

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #'preamble': '',
    # Latex figure (float) alignment
    #'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "bluesky.tex", "bluesky Documentation", "Brookhaven National Lab", "manual"),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "bluesky", "bluesky Documentation", [author], 1)]

# If true, show URL addresses after external links.
# man_show_urls = False


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "bluesky",
        "bluesky Documentation",
        author,
        "bluesky",
        "One line description of project.",
        "Miscellaneous",
    ),
]

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
# texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
# texinfo_no_detailmenu = False

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
    "ophyd": ("https://blueskyproject.io/ophyd/", None),
    "event-model": ("https://blueskyproject.io/event-model/main", None),
}

sys.path.append(os.path.abspath("examples"))



import threading
import time
import time as ttime
from functools import partial

import pytest

from bluesky import Msg
from bluesky.preprocessors import suspend_wrapper
from bluesky.run_engine import RunEngineInterrupted
from bluesky.suspenders import (
    SuspendBoolHigh,
    SuspendBoolLow,
    SuspendCeil,
    SuspendFloor,
    SuspendInBand,
    SuspendOutBand,
    SuspendWhenOutsideBand,
)
from bluesky.tests.utils import MsgCollector

from .utils import _fabricate_asycio_event


@pytest.mark.parametrize(
    "klass,sc_args,start_val,fail_val,resume_val,wait_time",
    [
        (SuspendBoolHigh, (), 0, 1, 0, 0.2),
        (SuspendBoolLow, (), 1, 0, 1, 0.2),
        (SuspendFloor, (0.5,), 1, 0, 1, 0.2),
        (SuspendCeil, (0.5,), 0, 1, 0, 0.2),
        (SuspendWhenOutsideBand, (0.5, 1.5), 1, 0, 1, 0.2),
        ((SuspendInBand, True), (0.5, 1.5), 1, 0, 1, 0.2),  # renamed to WhenOutsideBand
        ((SuspendOutBand, True), (0.5, 1.5), 0, 1, 0, 0.2),
    ],
)  # deprecated
def test_suspender(klass, sc_args, start_val, fail_val, resume_val, wait_time, RE, hw):
    sig = hw.bool_sig
    try:
        klass, deprecated = klass
    except TypeError:
        deprecated = False
    if deprecated:
        with pytest.warns(UserWarning):
            my_suspender = klass(sig, *sc_args, sleep=wait_time)
    else:
        my_suspender = klass(sig, *sc_args, sleep=wait_time)
    my_suspender.install(RE)

    def putter(val):
        sig.put(val)

    # make sure we start at good value!
    putter(start_val)
    # dumb scan
    scan = [Msg("checkpoint"), Msg("sleep", None, 0.2)]
    RE(scan)
    # paranoid
    assert RE.state == "idle"

    start = ttime.time()
    # queue up fail and resume conditions
    threading.Timer(0.1, putter, (fail_val,)).start()
    threading.Timer(0.5, putter, (resume_val,)).start()
    # start the scan
    RE(scan)
    stop = ttime.time()
    # assert we waited at least 2 seconds + the settle time
    delta = stop - start
    print(delta)
    assert delta > 0.5 + wait_time + 0.2


def test_pretripped(RE, hw):
    "Tests if suspender is tripped before __call__"
    sig = hw.bool_sig
    scan = [Msg("checkpoint")]
    msg_lst = []
    sig.put(1)

    def accum(msg):
        msg_lst.append(msg)

    susp = SuspendBoolHigh(sig)

    RE.install_suspender(susp)
    threading.Timer(1, sig.put, (0,)).start()
    RE.msg_hook = accum
    RE(scan)

    assert len(msg_lst) == 2
    assert ["wait_for", "checkpoint"] == [m[0] for m in msg_lst]


@pytest.mark.parametrize(
    "pre_plan,post_plan,expected_list",
    [
        (
            [Msg("null")],
            None,
            [
                "checkpoint",
                "sleep",
                "_start_suspender",
                "rewindable",
                "null",
                "wait_for",
                "_resume_from_suspender",
                "rewindable",
                "sleep",
            ],
        ),
        (
            None,
            [Msg("null")],
            [
                "checkpoint",
                "sleep",
                "_start_suspender",
                "rewindable",
                "wait_for",
                "_resume_from_suspender",
                "null",
                "rewindable",
                "sleep",
            ],
        ),
        (
            [Msg("null")],
            [Msg("null")],
            [
                "checkpoint",
                "sleep",
                "_start_suspender",
                "rewindable",
                "null",
                "wait_for",
                "_resume_from_suspender",
                "null",
                "rewindable",
                "sleep",
            ],
        ),
        (
            lambda: [Msg("null")],
            lambda: [Msg("null")],
            [
                "checkpoint",
                "sleep",
                "_start_suspender",
                "rewindable",
                "null",
                "wait_for",
                "_resume_from_suspender",
                "null",
                "rewindable",
                "sleep",
            ],
        ),
    ],
)
def test_pre_suspend_plan(RE, pre_plan, post_plan, expected_list, hw):
    sig = hw.bool_sig
    scan = [Msg("checkpoint"), Msg("sleep", None, 0.2)]
    msg_lst = []
    sig.put(0)

    def accum(msg):
        msg_lst.append(msg)

    susp = SuspendBoolHigh(sig, pre_plan=pre_plan, post_plan=post_plan)

    RE.install_suspender(susp)
    threading.Timer(0.1, sig.put, (1,)).start()
    threading.Timer(1, sig.put, (0,)).start()
    RE.msg_hook = accum
    RE(scan)

    assert len(msg_lst) == len(expected_list)
    assert expected_list == [m[0] for m in msg_lst]

    RE.remove_suspender(susp)
    RE(scan)
    assert susp.RE is None

    RE.install_suspender(susp)
    RE.clear_suspenders()
    assert susp.RE is None
    assert not RE.suspenders


def test_pause_from_suspend(RE, hw):
    "Tests what happens when a pause is requested from a suspended state"
    sig = hw.bool_sig
    scan = [Msg("checkpoint")]
    msg_lst = []
    sig.put(1)

    def accum(msg):
        msg_lst.append(msg)

    susp = SuspendBoolHigh(sig)

    RE.install_suspender(susp)
    threading.Timer(1, RE.request_pause).start()
    threading.Timer(2, sig.put, (0,)).start()
    RE.msg_hook = accum
    with pytest.raises(RunEngineInterrupted):
        RE(scan)
    assert [m[0] for m in msg_lst] == ["wait_for"]
    RE.resume()
    assert ["wait_for", "wait_for", "checkpoint"] == [m[0] for m in msg_lst]


def test_deferred_pause_from_suspend(RE, hw):
    "Tests what happens when a soft pause is requested from a suspended state"
    sig = hw.bool_sig
    scan = [Msg("checkpoint"), Msg("null")]
    msg_lst = []
    sig.put(1)

    def accum(msg):
        print(msg)
        msg_lst.append(msg)

    susp = SuspendBoolHigh(sig)

    RE.install_suspender(susp)
    threading.Timer(1, RE.request_pause, (True,)).start()
    threading.Timer(4, sig.put, (0,)).start()
    RE.msg_hook = accum
    with pytest.raises(RunEngineInterrupted):
        RE(scan)
    assert [m[0] for m in msg_lst] == ["wait_for", "checkpoint"]
    RE.resume()
    assert ["wait_for", "checkpoint", "null"] == [m[0] for m in msg_lst]


def test_unresumable_suspend_fail(RE):
    "Tests what happens when a soft pause is requested from a suspended state"

    scan = [Msg("clear_checkpoint"), Msg("sleep", None, 2)]
    m_coll = MsgCollector()
    RE.msg_hook = m_coll

    ev = _fabricate_asycio_event(RE.loop)
    threading.Timer(0.1, partial(RE.request_suspend, fut=ev.wait)).start()
    threading.Timer(1, ev.set).start()
    start = time.time()
    with pytest.raises(RunEngineInterrupted):
        RE(scan)
    stop = time.time()
    assert 0.1 < stop - start < 1


def test_suspender_plans(RE, hw):
    "Tests that the suspenders can be installed via Msg"
    sig = hw.bool_sig
    my_suspender = SuspendBoolHigh(sig, sleep=0.2)

    def putter(val):
        sig.put(val)

    putter(0)

    # Do the messages work?
    RE([Msg("install_suspender", None, my_suspender)])
    assert my_suspender in RE.suspenders
    RE([Msg("remove_suspender", None, my_suspender)])
    assert my_suspender not in RE.suspenders

    # Can we call both in a plan?
    RE([Msg("install_suspender", None, my_suspender), Msg("remove_suspender", None, my_suspender)])

    scan = [Msg("checkpoint"), Msg("sleep", None, 0.2)]

    # No suspend scan: does the wrapper error out?
    start = ttime.time()
    RE(suspend_wrapper(scan, my_suspender))
    stop = ttime.time()
    delta = stop - start
    assert delta < 0.9

    # Suspend scan
    start = ttime.time()
    threading.Timer(0.1, putter, (1,)).start()
    threading.Timer(0.5, putter, (0,)).start()
    RE(suspend_wrapper(scan, my_suspender))
    stop = ttime.time()
    delta = stop - start
    assert delta > 0.9

    # Did we clean up?
    start = ttime.time()
    threading.Timer(0.1, putter, (1,)).start()
    threading.Timer(0.5, putter, (0,)).start()
    RE(scan)
    stop = ttime.time()
    delta = stop - start
    assert delta < 0.9



from enum import Enum

import pytest

from bluesky._vendor.super_state_machine import utils


class StatesEnum(Enum):
    ONE = "one"
    TWO = "two"
    THREE = "three"
    FOUR = "four"


class OtherEnum(Enum):
    ONE = "one"


class NonUniqueEnum(Enum):
    ONE = "one"
    TWO = "two"
    THREE = "one"


class CollidingEnum(Enum):
    OPEN = "open"
    OPENING = "opening"
    CLOSE = "close"
    CLOSED = "closed"


def test_translator():
    trans = utils.EnumValueTranslator(StatesEnum)
    assert trans.translate("one") == StatesEnum.ONE
    assert trans.translate("two") == StatesEnum.TWO
    assert trans.translate("three") == StatesEnum.THREE
    assert trans.translate("four") == StatesEnum.FOUR


def test_translator_for_wrong_values():
    trans = utils.EnumValueTranslator(StatesEnum)
    with pytest.raises(ValueError):
        trans.translate("a")
    with pytest.raises(ValueError):
        trans.translate("x")
    with pytest.raises(ValueError):
        trans.translate("threex")
    with pytest.raises(ValueError):
        trans.translate("threx")
    with pytest.raises(ValueError):
        trans.translate("fake")


def test_translator_for_enum_value():
    trans = utils.EnumValueTranslator(StatesEnum)
    assert trans.translate(StatesEnum.ONE) is StatesEnum.ONE
    assert trans.translate(StatesEnum.TWO) is StatesEnum.TWO
    with pytest.raises(ValueError):
        trans.translate(OtherEnum.ONE)


def test_translator_doesnt_accept_non_unique_enums():
    with pytest.raises(ValueError):
        utils.EnumValueTranslator(NonUniqueEnum)



import gc
import os
import signal
import threading
import time
from subprocess import run

import multiprocess
import numpy as np
import pytest
from event_model import sanitize_doc

from bluesky import Msg
from bluesky.callbacks.zmq import Proxy, Publisher, RemoteDispatcher
from bluesky.plans import count
from bluesky.tests import uses_os_kill_sigint


def test_proxy_script():
    p = run(["bluesky-0MQ-proxy", "-h"])
    assert p.returncode == 0


def test_zmq(RE, hw):
    # COMPONENT 1
    # Run a 0MQ proxy on a separate process.
    def start_proxy():
        Proxy(5567, 5568).start()

    proxy_proc = multiprocess.Process(target=start_proxy, daemon=True)
    proxy_proc.start()
    time.sleep(5)  # Give this plenty of time to start up.

    # COMPONENT 2
    # Run a Publisher and a RunEngine in this main process.

    p = Publisher("127.0.0.1:5567")  # noqa
    RE.subscribe(p)

    # COMPONENT 3
    # Run a RemoteDispatcher on another separate process. Pass the documents
    # it receives over a Queue to this process, so we can count them for our
    # test.

    def make_and_start_dispatcher(queue):
        def put_in_queue(name, doc):
            print("putting ", name, "in queue")
            queue.put((name, doc))

        d = RemoteDispatcher("127.0.0.1:5568")
        d.subscribe(put_in_queue)
        print("REMOTE IS READY TO START")
        d.loop.call_later(9, d.stop)
        d.start()

    queue = multiprocess.Queue()
    dispatcher_proc = multiprocess.Process(target=make_and_start_dispatcher, daemon=True, args=(queue,))
    dispatcher_proc.start()
    time.sleep(5)  # As above, give this plenty of time to start.

    # Generate two documents. The Publisher will send them to the proxy
    # device over 5567, and the proxy will send them to the
    # RemoteDispatcher over 5568. The RemoteDispatcher will push them into
    # the queue, where we can verify that they round-tripped.

    local_accumulator = []

    def local_cb(name, doc):
        local_accumulator.append((name, doc))

    # Check that numpy stuff is sanitized by putting some in the start doc.
    md = {"stuff": {"nested": np.array([1, 2, 3])}, "scalar_stuff": np.float64(3), "array_stuff": np.ones((3, 3))}

    # RE([Msg('open_run', **md), Msg('close_run')], local_cb)
    RE(count([hw.det]), local_cb, **md)
    time.sleep(1)

    # Get the two documents from the queue (or timeout --- test will fail)
    remote_accumulator = []
    for i in range(len(local_accumulator)):  # noqa: B007
        remote_accumulator.append(queue.get(timeout=2))
    p.close()
    proxy_proc.terminate()
    dispatcher_proc.terminate()
    proxy_proc.join()
    dispatcher_proc.join()
    ra = sanitize_doc(remote_accumulator)
    la = sanitize_doc(local_accumulator)
    assert ra == la

    gc.collect()
    gc.collect()


@uses_os_kill_sigint
def test_zmq_proxy_blocks_sigint_exits():
    # The test `test_zmq` runs Proxy and RemoteDispatcher in a separate
    # process, which coverage misses.

    def delayed_sigint(delay):
        time.sleep(delay)
        os.kill(os.getpid(), signal.SIGINT)

    proxy = Proxy(5567, 5568)
    assert not proxy.closed
    threading.Thread(target=delayed_sigint, args=(1,)).start()
    try:
        proxy.start()
        # delayed_sigint stops the proxy
    except KeyboardInterrupt:
        ...
    assert proxy.closed
    with pytest.raises(RuntimeError):
        proxy.start()

    proxy = Proxy()  # random port
    threading.Thread(target=delayed_sigint, args=(1,)).start()
    try:
        proxy.start()
        # delayed_sigint stops the proxy
    except KeyboardInterrupt:
        ...
    assert proxy.closed
    repr(proxy)
    gc.collect()
    gc.collect()


@pytest.mark.parametrize("host", ["localhost:5555", ("localhost", 5555)])
def test_zmq_RD_ports_spec(host):
    # test that two ways of specifying address are equivalent
    d = RemoteDispatcher(host)
    assert d.address == ("localhost", 5555)
    assert d._socket is None
    assert d._context is None
    assert not d.closed
    d.stop()
    assert d._socket is None
    assert d._context is None
    assert d.closed
    del d


def test_zmq_no_RE(RE):
    # COMPONENT 1
    # Run a 0MQ proxy on a separate process.
    def start_proxy():
        Proxy(5567, 5568).start()

    proxy_proc = multiprocess.Process(target=start_proxy, daemon=True)
    proxy_proc.start()
    time.sleep(5)  # Give this plenty of time to start up.

    # COMPONENT 2
    # Run a Publisher and a RunEngine in this main process.

    p = Publisher("127.0.0.1:5567")  # noqa

    # COMPONENT 3
    # Run a RemoteDispatcher on another separate process. Pass the documents
    # it receives over a Queue to this process, so we can count them for our
    # test.

    def make_and_start_dispatcher(queue):
        def put_in_queue(name, doc):
            print("putting ", name, "in queue")
            queue.put((name, doc))

        d = RemoteDispatcher("127.0.0.1:5568")
        d.subscribe(put_in_queue)
        print("REMOTE IS READY TO START")
        d.loop.call_later(9, d.stop)
        d.start()

    queue = multiprocess.Queue()
    dispatcher_proc = multiprocess.Process(target=make_and_start_dispatcher, daemon=True, args=(queue,))
    dispatcher_proc.start()
    time.sleep(5)  # As above, give this plenty of time to start.

    # Generate two documents. The Publisher will send them to the proxy
    # device over 5567, and the proxy will send them to the
    # RemoteDispatcher over 5568. The RemoteDispatcher will push them into
    # the queue, where we can verify that they round-tripped.

    local_accumulator = []

    def local_cb(name, doc):
        local_accumulator.append((name, doc))

    RE([Msg("open_run"), Msg("close_run")], local_cb)

    # This time the Publisher isn't attached to an RE. Send the documents
    # manually. (The idea is, these might have come from a Broker instead...)
    for name, doc in local_accumulator:
        p(name, doc)
    time.sleep(1)

    # Get the two documents from the queue (or timeout --- test will fail)
    remote_accumulator = []
    for i in range(2):  # noqa: B007
        remote_accumulator.append(queue.get(timeout=2))
    p.close()
    proxy_proc.terminate()
    dispatcher_proc.terminate()
    proxy_proc.join()
    dispatcher_proc.join()
    ra = sanitize_doc(remote_accumulator)
    la = sanitize_doc(local_accumulator)
    assert ra == la


def test_zmq_no_RE_newserializer(RE):
    cloudpickle = pytest.importorskip("cloudpickle")

    # COMPONENT 1
    # Run a 0MQ proxy on a separate process.
    def start_proxy():
        Proxy(5567, 5568).start()

    proxy_proc = multiprocess.Process(target=start_proxy, daemon=True)
    proxy_proc.start()
    time.sleep(5)  # Give this plenty of time to start up.

    # COMPONENT 2
    # Run a Publisher and a RunEngine in this main process.
    p = Publisher("127.0.0.1:5567", serializer=cloudpickle.dumps)  # noqa

    # COMPONENT 3
    # Run a RemoteDispatcher on another separate process. Pass the documents
    # it receives over a Queue to this process, so we can count them for our
    # test.
    def make_and_start_dispatcher(queue):
        def put_in_queue(name, doc):
            print("putting ", name, "in queue")
            queue.put((name, doc))

        d = RemoteDispatcher("127.0.0.1:5568", deserializer=cloudpickle.loads)
        d.subscribe(put_in_queue)
        print("REMOTE IS READY TO START")
        d.loop.call_later(9, d.stop)
        d.start()

    queue = multiprocess.Queue()
    dispatcher_proc = multiprocess.Process(target=make_and_start_dispatcher, daemon=True, args=(queue,))
    dispatcher_proc.start()
    time.sleep(5)  # As above, give this plenty of time to start.

    # Generate two documents. The Publisher will send them to the proxy
    # device over 5567, and the proxy will send them to the
    # RemoteDispatcher over 5568. The RemoteDispatcher will push them into
    # the queue, where we can verify that they round-tripped.

    local_accumulator = []

    def local_cb(name, doc):
        local_accumulator.append((name, doc))

    RE([Msg("open_run"), Msg("close_run")], local_cb)

    # This time the Publisher isn't attached to an RE. Send the documents
    # manually. (The idea is, these might have come from a Broker instead...)
    for name, doc in local_accumulator:
        p(name, doc)
    time.sleep(1)

    # Get the two documents from the queue (or timeout --- test will fail)
    remote_accumulator = []
    for i in range(2):  # noqa: B007
        remote_accumulator.append(queue.get(timeout=2))
    p.close()
    proxy_proc.terminate()
    dispatcher_proc.terminate()
    proxy_proc.join()
    dispatcher_proc.join()
    ra = sanitize_doc(remote_accumulator)
    la = sanitize_doc(local_accumulator)
    assert ra == la


def test_zmq_prefix(RE, hw):
    # COMPONENT 1
    # Run a 0MQ proxy on a separate process.
    def start_proxy():
        Proxy(5567, 5568).start()

    proxy_proc = multiprocess.Process(target=start_proxy, daemon=True)
    proxy_proc.start()
    time.sleep(5)  # Give this plenty of time to start up.

    # COMPONENT 2
    # Run a Publisher and a RunEngine in this main process.
    p = Publisher("127.0.0.1:5567", prefix=b"sb")  # noqa
    p2 = Publisher("127.0.0.1:5567", prefix=b"not_sb")  # noqa
    RE.subscribe(p)
    RE.subscribe(p2)

    # COMPONENT 3
    # Run a RemoteDispatcher on another separate process. Pass the documents
    # it receives over a Queue to this process, so we can count them for our
    # test.

    def make_and_start_dispatcher(queue):
        def put_in_queue(name, doc):
            print("putting ", name, "in queue")
            queue.put((name, doc))

        d = RemoteDispatcher("127.0.0.1:5568", prefix=b"sb")
        d.subscribe(put_in_queue)
        print("REMOTE IS READY TO START")
        d.loop.call_later(9, d.stop)
        d.start()

    queue = multiprocess.Queue()
    dispatcher_proc = multiprocess.Process(target=make_and_start_dispatcher, daemon=True, args=(queue,))
    dispatcher_proc.start()
    time.sleep(5)  # As above, give this plenty of time to start.

    # Generate two documents. The Publisher will send them to the proxy
    # device over 5567, and the proxy will send them to the
    # RemoteDispatcher over 5568. The RemoteDispatcher will push them into
    # the queue, where we can verify that they round-tripped.

    local_accumulator = []

    def local_cb(name, doc):
        local_accumulator.append((name, doc))

    # Check that numpy stuff is sanitized by putting some in the start doc.
    md = {"stuff": {"nested": np.array([1, 2, 3])}, "scalar_stuff": np.float64(3), "array_stuff": np.ones((3, 3))}

    # RE([Msg('open_run', **md), Msg('close_run')], local_cb)
    RE(count([hw.det]), local_cb, **md)
    time.sleep(1)

    # Get the two documents from the queue (or timeout --- test will fail)
    remote_accumulator = []
    for i in range(len(local_accumulator)):  # noqa: B007
        remote_accumulator.append(queue.get(timeout=2))
    p.close()
    p2.close()
    proxy_proc.terminate()
    dispatcher_proc.terminate()
    proxy_proc.join()
    dispatcher_proc.join()
    ra = sanitize_doc(remote_accumulator)
    la = sanitize_doc(local_accumulator)
    assert ra == la



import jsonschema
import pytest
from event_model import DocumentNames, schema_validators

from bluesky.run_engine import Msg
from bluesky.utils import new_uid


def simple_scan(motor):
    yield Msg("open_run")
    yield Msg("set", motor, 5)
    yield Msg("read", motor)
    yield Msg("close_run")


def test_custom_metadata(RE, hw):
    def assert_lion(name, doc):
        assert "animal" in doc
        assert doc["animal"] == "lion"

    RE(simple_scan(hw.motor), {"start": assert_lion}, animal="lion")
    # Note: Because assert_lion is processed on the main thread, it can
    # fail the test. I checked by writing a failing version of it.  - D.A.


def test_dots_not_allowed_in_keys():
    doc = {"time": 0, "uid": new_uid()}
    schema_validators[DocumentNames.start].validate(doc)
    # Add a legal key.
    doc.update({"b": "c"})
    schema_validators[DocumentNames.start].validate(doc)
    # Now add illegal key.
    doc.update({"b.": "c"})
    with pytest.raises(jsonschema.ValidationError):
        schema_validators[DocumentNames.start].validate(doc)

    doc = {
        "time": 0,
        "uid": new_uid(),
        "data_keys": {"a": {"source": "", "dtype": "number", "shape": []}},
        "run_start": new_uid(),
    }
    schema_validators[DocumentNames.descriptor].validate(doc)
    # Add a legal key.
    doc.update({"b": "c"})
    schema_validators[DocumentNames.descriptor].validate(doc)
    # Now add illegal key.
    doc.update({"b.c": "d"})
    with pytest.raises(jsonschema.ValidationError):
        schema_validators[DocumentNames.descriptor].validate(doc)

    doc = {"time": 0, "uid": new_uid(), "exit_status": "success", "reason": "", "run_start": new_uid()}
    schema_validators[DocumentNames.stop].validate(doc)
    # Add a legal key.
    doc.update({"b": "c"})
    schema_validators[DocumentNames.stop].validate(doc)
    # Now add illegal key.
    doc.update({".b": "c"})
    with pytest.raises(jsonschema.ValidationError):
        schema_validators[DocumentNames.stop].validate(doc)



from collections.abc import Generator, Sequence
from itertools import dropwhile
from time import time
from typing import (
    Any,
    Callable,
    Literal,
    Optional,
    Union,
    cast,
)
from warnings import warn

from bluesky.log import logger as LOGGER
from bluesky.preprocessors import print_summary_wrapper
from bluesky.run_engine import call_in_bluesky_event_loop, in_bluesky_event_loop
from bluesky.utils import Msg, maybe_await

from .protocols import Checkable, Readable, Reading

END = "end"


def plot_raster_path(plan, x_motor, y_motor, ax=None, probe_size=None, lw=2):
    """Plot the raster path for this plan

    Parameters
    ----------
    plan : iterable
       Must yield `Msg` objects and not be a co-routine

    x_motor, y_motor : str
       Names of the x and y motors

    ax : matplotlib.axes.Axes
       The axes to plot to, if none, make new figure + axes

    probe_size : float, optional
       If not None, use as radius of probe (in same units as motor positions)

    lw : float, optional
        Width of lines drawn between points
    """
    import matplotlib.pyplot as plt
    from matplotlib import collections as mcollections
    from matplotlib import patches as mpatches

    if ax is None:
        ax = plt.subplots()[1]
    ax.set_aspect("equal")

    cur_x = cur_y = None
    traj = []
    for msg in plan:
        cmd = msg.command
        if cmd == "set":
            if msg.obj.name == x_motor:
                cur_x = msg.args[0]
            if msg.obj.name == y_motor:
                cur_y = msg.args[0]
        elif cmd == "save":
            traj.append((cur_x, cur_y))

    x, y = zip(*traj)
    (path,) = ax.plot(x, y, marker="", linestyle="-", lw=lw)
    ax.set_xlabel(x_motor)
    ax.set_ylabel(y_motor)
    if probe_size is None:
        read_points = ax.scatter(x, y, marker="o", lw=lw)
    else:
        circles = [mpatches.Circle((_x, _y), probe_size, facecolor="black", alpha=0.5) for _x, _y in traj]

        read_points = mcollections.PatchCollection(circles, match_original=True)
        ax.add_collection(read_points)
    return {"path": path, "events": read_points}


def summarize_plan(plan):
    """Print summary of plan

    Prints a minimal version of the plan, showing only moves and
    where events are created.

    Parameters
    ----------
    plan : iterable
        Must yield `Msg` objects
    """
    for msg in print_summary_wrapper(plan):  # noqa: B007
        ...


print_summary = summarize_plan  # back-compat


def check_limits(plan):
    """Run check_limits_async in the RE"""
    if in_bluesky_event_loop():
        raise RuntimeError("Can't call check_limits() from within RE, use await check_limits_async() instead")
    call_in_bluesky_event_loop(check_limits_async(plan))


async def check_limits_async(plan):
    """
    Check that a plan will not move devices outside of their limits.

    Parameters
    ----------
    plan : iterable
        Must yield `Msg` objects
    """
    ignore = []
    for msg in plan:
        obj = msg.obj
        if msg.command == "set" and obj not in ignore:
            if isinstance(obj, Checkable):
                await maybe_await(obj.check_value(msg.args[0]))
            else:
                warn(  # noqa: B028
                    f"{obj.name} has no check_value() method to check if {msg.args[0]} is within its limits."
                )
                ignore.append(obj)


class RunEngineSimulator:
    """Helps test a Bluesky plan by recording bluesky messages and
     injecting responses according to the bluesky Message Protocol.

    See bluesky docs for details of the message protocol.

    Basic usage consists of
    1) Registering various handlers to respond to anticipated messages in the experiment plan and fire
     any needed callbacks.
    2) Calling simulate_plan()
    3) Examining the returned message list and making asserts against them.

    Attributes
    ----------
    return_value
        The return value of the most recently executed plan.

    Examples
    --------

    >>> def my_plan():
    ...     pass

    >>> sim = RunEngineSimulator()
    >>> messages = sim.simulate_plan(my_plan())
    >>>
    """

    GROUP_ANY = "any"

    def __init__(self) -> None:
        self.message_handlers: list[_MessageHandler] = []
        self.callbacks: dict[int, tuple[Callable[[str, dict], None], str]] = {}
        self.next_callback_token: int = 0
        self.return_value: Any = None

    def add_handler_for_callback_subscribes(self):
        """Add a handler that registers all the callbacks from subscribe messages so we can call them later.

        You will need to call this as one of the first things if you wish to fire callbacks from the simulator.
        """
        self.message_handlers.append(
            _MessageHandler(
                lambda msg: msg.command == "subscribe",
                lambda msg: self._add_callback(msg.args),
            )
        )

    def add_handler(
        self,
        commands: Union[str, Sequence[str]],
        handler: Callable[[Msg], object],
        msg_filter: Optional[Union[str, Callable[[Msg], bool]]] = None,
        index: Union[int, Literal["end"]] = 0,
    ):
        """Add the specified handler for a particular message.

        Parameters
        ----------
        commands
            The command name for the message as defined in bluesky Message Protocol, or a sequence if
            more than one matches.
        handler
            A lambda that accepts a Msg and returns an object; the object is sent to the current yield
            statement in the generator, and is used when reading values from devices, the structure of the
            object depends on device hinting.
        msg_filter
            This can either be a string corresponding to the Msg obj.name property to match, or it can
            be a predicate on the Msg which returns a bool. The default is None which will match all
            messages.
        index
            An optional integer indicating where to insert the handler, the default is 0 which
            prepends it so that newer handlers override older ones. Specify END to append the handler.

        """
        if isinstance(commands, str):
            commands = [commands]

        self.message_handlers.insert(
            cast(int, index if index != END else len(self.message_handlers)),
            _MessageHandler(
                lambda msg: msg.command in commands
                and (
                    msg_filter is None
                    or (callable(msg_filter) and msg_filter(msg))
                    or (msg.obj and msg.obj.name == msg_filter)
                ),
                handler,
            ),
        )

    def add_read_handler_for(self, obj: Readable, value: Optional[Any]):
        """
        Convenience method to register a handler to return a result from a
        single-valued 'read' command.
        The behaviour is equivalent to
        add_read_handler_for_multiple(obj, **{obj.name: value})

        Examples
        --------
        In order to exercise a plan containing the following:

        >>> def trigger_and_return_pin_tip():
        ...     yield from bps.trigger(pin_tip)
        ...     tip_pos = yield from bps.rd(pin_tip.triggered_tip)
        ...     return tip_pos

        >>> sim = RunEngineSimulator()
        >>> sim.add_read_handler("pin_tip_detection-triggered_tip", (8, 5))
        >>> msgs = sim.simulate_plan(trigger_and_return_pin_tip())
        >>> assert sim.return_value == (8, 5)
        """
        self.add_read_handler_for_multiple(obj, **{obj.name: value})

    def add_read_handler_for_multiple(self, obj: Readable, **kwargs):
        """
        Convenience method to register a handler to return a result from a multi-valued
        'read' command.

        Parameters
        ----------
        obj
           The object to intercept reads for
        kwargs
            The handler will return a dict with the corresponding key-value pairs.
            If the value is a dict the result of the read command for the key is the supplied dict.
            Otherwise, the value of the read command for the key is
            a value of Reading(value=value, timestamp=time())

        Examples
        --------
        In order to exercise a plan containing the following:

        >>> def read_all_values():
        >>>     data = yield from bps.read(hw.ab_det)
        >>>     return data
        >>> sim = RunEngineSimulator()
        >>> sim.add_read_handler_for_multiple(hw.ab_det, a=11, b=12)
        >>> sim.simulate_plan(read_all_values())
        >>> assert sim.return_value["a"]["value"] == 11
        >>> assert sim.return_value["b"]["value"] == 12
        >>> assert isclose(sim.return_value["a"]["timestamp"], time(), abs_tol=1)
        >>> assert sim.return_value["b"]["timestamp"] == 1717071719
        """

        def handle_read(_):
            return {k: v if isinstance(v, dict) else Reading(value=v, timestamp=time()) for k, v in kwargs.items()}

        self.add_handler("read", handle_read, obj.name)

    def add_wait_handler(self, handler: Callable[[Msg], None], group: str = GROUP_ANY) -> None:
        """Add a wait handler for a particular message.

        Parameters
        ----------
        handler
            a lambda that accepts a Msg, use this to execute any code that simulates something that's
            supposed to complete when a group finishes
        group
            name of the group to wait for, default is GROUP_ANY which matches them all

        Examples
        --------
        In order to exercise the following plan:
        >>> def close_shutter():
        ...     yield from bps.abs_set(detector_motion.shutter, 1, group="my_group")
        ...     yield from bps.wait("my_group")
        ...     done = yield from bps.rd("detector_motion_shutter")
        ...     assert done == 1

        >>> sim = RunEngineSimulator()
        ... def simulate_detector_motion():
        ...     sim.add_read_handler("detector_motion_shutter", 1)
        ...
        ... sim.add_wait_handler(simulate_detector_motion, "my_group")
        ... sim.simulate_plan(close_shutter())
        """
        self.add_handler(
            "wait",
            handler,
            lambda msg: (group == RunEngineSimulator.GROUP_ANY or msg.kwargs["group"] == group),
        )

    def add_callback_handler_for(
        self,
        command: str,
        document_name: str,
        document: dict,
        msg_filter: Optional[Callable[[Msg], bool]] = None,
    ):
        """Add a handler to fire a callback when a matching command is encountered.
        Equivalent to add_callback_for_multiple(command, [[(document_name, document)]], msg_filter)
        Parameters
        ----------
        command
            Name of the command to match.
        document_name
            Name of the document to match when used in single shot mode.
        document
            The document to fire in single shot mode.
        msg_filter
            Optional predicate on the message which triggers the callback
        """
        self.add_callback_handler_for_multiple(command, [[(document_name, document)]], msg_filter)

    def add_callback_handler_for_multiple(
        self,
        command: str,
        docs: Sequence[Sequence[tuple[str, dict]]],
        msg_filter: Optional[Callable[[Msg], bool]] = None,
    ):
        """Add a handler to fire callbacks in sequence when a matching command is encountered.

        Parameters
        ----------
        command
            Name of the command to match.
        docs
            Sequence of Sequence of document_name, document tuples
            On each receipt of a matching message, the sequence contained in the next element of `docs`
            will be iterated over and each (document_name, document) pair will be fired to subscribed callbacks
            in order.
        msg_filter
            Optional predicate on the message which triggers the callback
        """
        it = iter(docs)

        def handle_command(_):
            for name, doc in next(it):
                self.fire_callback(name, doc)

        self.add_handler(command, handle_command, msg_filter)

    def fire_callback(self, document_name, document) -> None:
        """Fire all the callbacks registered for this document type in order to simulate something happening.

        Parameters
        ----------
        document_name
            document name as defined in the Bluesky Message Protocol 'subscribe' call,
            all subscribers filtering on this document name will be called
        document
            the document to send
        """
        for callback_func, callback_docname in self.callbacks.values():
            if callback_docname == "all" or callback_docname == document_name:
                callback_func(document_name, document)

    def simulate_plan(self, gen: Generator[Msg, Any, Any]) -> list[Msg]:
        """Simulate the RunEngine executing the plan.

        After executing the plan return_value is populated with the return value of the plan.

        Parameters
        ----------
        gen
            the generator function that executes the plan

        Returns
        -------
        list[Msg]
            list of the messages generated by the plan
        """
        messages = []
        send_value = None
        try:
            while msg := gen.send(send_value):
                send_value = None
                messages.append(msg)
                LOGGER.debug("<%s", msg)
                if handler := next((h for h in self.message_handlers if h.predicate(msg)), None):
                    send_value = handler.runnable(msg)

                if send_value:
                    LOGGER.debug(f">send {send_value}")
        except StopIteration as e:
            self.return_value = e.value
        return messages

    def _add_callback(self, msg_args):
        self.callbacks[self.next_callback_token] = msg_args
        self.next_callback_token += 1


def assert_message_and_return_remaining(
    messages: list[Msg],
    predicate: Callable[[Msg], bool],
    group: Optional[str] = None,
):
    """Find the next message matching the predicate, assert that we found it.

    Returns
    -------
    list[Msg]
        All the remaining messages starting from the matched message which is included as a
        convenience to capture information that may be used in subsequent calls.
    """

    def not_matching(msg: Msg) -> bool:
        return (group is not None and (msg.kwargs and msg.kwargs.get("group") != group)) or not predicate(msg)

    matched = list(dropwhile(not_matching, messages))
    assert matched, f"Nothing matched predicate {predicate}"
    return matched


class _MessageHandler:
    def __init__(self, p: Callable[[Msg], bool], r: Callable[[Msg], object]):
        self.predicate = p
        self.runnable = r
